import os
import time


class NetlinkError(Exception):
    '''
    Base netlink error
    '''

    def __init__(self, code, msg=None):
        msg = msg or os.strerror(code)
        super(NetlinkError, self).__init__(code, msg)
        self.code = code
        self.extra_code = 0


class NetlinkDecodeError(Exception):
    '''
    Base decoding error class.

    Incapsulates underlying error for the following analysis
    '''

    def __init__(self, exception):
        self.exception = exception


class NetlinkHeaderDecodeError(NetlinkDecodeError):
    '''
    The error occured while decoding a header
    '''

    pass


class NetlinkDataDecodeError(NetlinkDecodeError):
    '''
    The error occured while decoding the message fields
    '''

    pass


class NetlinkNLADecodeError(NetlinkDecodeError):
    '''
    The error occured while decoding NLA chain
    '''

    pass


class IPSetError(NetlinkError):
    '''
    Netlink error with IPSet special error codes.

    Messages are imported from errcode.c
    '''

    pass


class NetlinkDumpInterrupted(NetlinkError):
    '''
    Raised when NLM_F_DUMP_INTR is set in the flags.
    '''

    def __init__(self, code=-1, msg='dump interrupted'):
        super(NetlinkDumpInterrupted, self).__init__(code, msg)


class SkipInode(Exception):
    def __init__(self, code=0, msg=None):
        super(SkipInode, self).__init__(code, msg)
        self.code = code


class ChaoticException(Exception):
    def __init__(self):
        chaotic_id = str(time.time())
        super(ChaoticException, self).__init__(chaotic_id)
        self.chaotic_id = chaotic_id



import struct
from socket import AF_INET, inet_ntop, inet_pton

from pyroute2.common import basestring, hexdump

#
# IEEE = 802.3 Ethernet magic constants. The frame sizes omit
# the preamble and FCS/CRC (frame check sequence).
#

ETH_ALEN = 6  # Octets in one ethernet addr
ETH_HLEN = 14  # Total octets in header.
ETH_ZLEN = 60  # Min. octets in frame sans FCS
ETH_DATA_LEN = 1500  # Max. octets in payload
ETH_FRAME_LEN = 1514  # Max. octets in frame sans FCS
ETH_FCS_LEN = 4  # Octets in the FCS

#
# These are the defined Ethernet Protocol ID's.
#

ETH_P_LOOP = 0x0060  # Ethernet Loopback packet
ETH_P_PUP = 0x0200  # Xerox PUP packet
ETH_P_PUPAT = 0x0201  # Xerox PUP Addr Trans packet
ETH_P_IP = 0x0800  # Internet Protocol packet
ETH_P_X25 = 0x0805  # CCITT X.25
ETH_P_ARP = 0x0806  # Address Resolution packet
ETH_P_BPQ = 0x08FF  # G8BPQ AX.25 Ethernet Packet
# ^^^ [ NOT AN OFFICIALLY REGISTERED ID ]
ETH_P_IEEEPUP = 0x0A00  # Xerox IEEE802.3 PUP packet
ETH_P_IEEEPUPAT = 0x0A01  # Xerox IEEE802.3 PUP Addr Trans packet
ETH_P_DEC = 0x6000  # DEC Assigned proto
ETH_P_DNA_DL = 0x6001  # DEC DNA Dump/Load
ETH_P_DNA_RC = 0x6002  # DEC DNA Remote Console
ETH_P_DNA_RT = 0x6003  # DEC DNA Routing
ETH_P_LAT = 0x6004  # DEC LAT
ETH_P_DIAG = 0x6005  # DEC Diagnostics
ETH_P_CUST = 0x6006  # DEC Customer use
ETH_P_SCA = 0x6007  # DEC Systems Comms Arch
ETH_P_TEB = 0x6558  # Trans Ether Bridging
ETH_P_RARP = 0x8035  # Reverse Addr Res packet
ETH_P_ATALK = 0x809B  # Appletalk DDP
ETH_P_AARP = 0x80F3  # Appletalk AARP
ETH_P_8021Q = 0x8100  # = 802.1Q VLAN Extended Header
ETH_P_IPX = 0x8137  # IPX over DIX
ETH_P_IPV6 = 0x86DD  # IPv6 over bluebook
ETH_P_PAUSE = 0x8808  # IEEE Pause frames. See = 802.3 = 31B
ETH_P_SLOW = 0x8809  # Slow Protocol. See = 802.3ad = 43B
ETH_P_WCCP = 0x883E  # Web-cache coordination protocol
# defined in draft-wilson-wrec-wccp-v2-00.txt
ETH_P_PPP_DISC = 0x8863  # PPPoE discovery messages
ETH_P_PPP_SES = 0x8864  # PPPoE session messages
ETH_P_MPLS_UC = 0x8847  # MPLS Unicast traffic
ETH_P_MPLS_MC = 0x8848  # MPLS Multicast traffic
ETH_P_ATMMPOA = 0x884C  # MultiProtocol Over ATM
ETH_P_LINK_CTL = 0x886C  # HPNA, wlan link local tunnel
ETH_P_ATMFATE = 0x8884  # Frame-based ATM Transport over Ethernet

ETH_P_PAE = 0x888E  # Port Access Entity (IEEE = 802.1X)
ETH_P_AOE = 0x88A2  # ATA over Ethernet
ETH_P_8021AD = 0x88A8  # = 802.1ad Service VLAN
ETH_P_802_EX1 = 0x88B5  # = 802.1 Local Experimental = 1.
ETH_P_TIPC = 0x88CA  # TIPC
ETH_P_8021AH = 0x88E7  # = 802.1ah Backbone Service Tag
ETH_P_1588 = 0x88F7  # IEEE = 1588 Timesync
ETH_P_FCOE = 0x8906  # Fibre Channel over Ethernet
ETH_P_TDLS = 0x890D  # TDLS
ETH_P_FIP = 0x8914  # FCoE Initialization Protocol
ETH_P_QINQ1 = 0x9100  # deprecated QinQ VLAN
# ^^^ [ NOT AN OFFICIALLY REGISTERED ID ]
ETH_P_QINQ2 = 0x9200  # deprecated QinQ VLAN
# ^^^ [ NOT AN OFFICIALLY REGISTERED ID ]
ETH_P_QINQ3 = 0x9300  # deprecated QinQ VLAN
# ^^^ [ NOT AN OFFICIALLY REGISTERED ID ]
ETH_P_EDSA = 0xDADA  # Ethertype DSA
# ^^^ [ NOT AN OFFICIALLY REGISTERED ID ]
ETH_P_AF_IUCV = 0xFBFB  # IBM af_iucv
# ^^^ [ NOT AN OFFICIALLY REGISTERED ID ]

#
# Non DIX types. Won't clash for = 1500 types.
#

ETH_P_802_3 = 0x0001  # Dummy type for = 802.3 frames
ETH_P_AX25 = 0x0002  # Dummy protocol id for AX.25
ETH_P_ALL = 0x0003  # Every packet (be careful!!!)
ETH_P_802_2 = 0x0004  # = 802.2 frames
ETH_P_SNAP = 0x0005  # Internal only
ETH_P_DDCMP = 0x0006  # DEC DDCMP: Internal only
ETH_P_WAN_PPP = 0x0007  # Dummy type for WAN PPP frames*/
ETH_P_PPP_MP = 0x0008  # Dummy type for PPP MP frames
ETH_P_LOCALTALK = 0x0009  # Localtalk pseudo type
ETH_P_CAN = 0x000C  # Controller Area Network
ETH_P_PPPTALK = 0x0010  # Dummy type for Atalk over PPP*/
ETH_P_TR_802_2 = 0x0011  # = 802.2 frames
ETH_P_MOBITEX = 0x0015  # Mobitex (kaz@cafe.net)
ETH_P_CONTROL = 0x0016  # Card specific control frames
ETH_P_IRDA = 0x0017  # Linux-IrDA
ETH_P_ECONET = 0x0018  # Acorn Econet
ETH_P_HDLC = 0x0019  # HDLC frames
ETH_P_ARCNET = 0x001A  # = 1A for ArcNet :-)
ETH_P_DSA = 0x001B  # Distributed Switch Arch.
ETH_P_TRAILER = 0x001C  # Trailer switch tagging
ETH_P_PHONET = 0x00F5  # Nokia Phonet frames
ETH_P_IEEE802154 = 0x00F6  # IEEE802.15.4 frame
ETH_P_CAIF = 0x00F7  # ST-Ericsson CAIF protocol


class msg(dict):
    buf = None
    data_len = None
    fields = ()
    _fields_names = ()
    types = {
        'uint8': 'B',
        'uint16': 'H',
        'uint32': 'I',
        'be16': '>H',
        'ip4addr': {
            'format': '4s',
            'decode': lambda x: inet_ntop(AF_INET, x),
            'encode': lambda x: [inet_pton(AF_INET, x)],
        },
        'l2addr': {
            'format': '6B',
            'decode': lambda x: ':'.join(['%x' % i for i in x]),
            'encode': lambda x: [int(i, 16) for i in x.split(':')],
        },
        'l2paddr': {
            'format': '6B10s',
            'decode': lambda x: ':'.join(['%x' % i for i in x[:6]]),
            'encode': lambda x: [int(i, 16) for i in x.split(':')]
            + [10 * b'\x00'],
        },
    }

    def __init__(self, content=None, buf=b'', offset=0, value=None):
        content = content or {}
        dict.__init__(self, content)
        self.buf = buf
        self.offset = offset
        self.value = value
        self._register_fields()

    def _register_fields(self):
        self._fields_names = tuple([x[0] for x in self.fields])

    def _get_routine(self, mode, fmt):
        fmt = self.types.get(fmt, fmt)
        if isinstance(fmt, dict):
            return (fmt['format'], fmt.get(mode, lambda x: x))
        else:
            return (fmt, lambda x: x)

    def reset(self):
        self.buf = b''

    def decode(self):
        self._register_fields()
        for field in self.fields:
            name, sfmt = field[:2]
            fmt, routine = self._get_routine('decode', sfmt)
            size = struct.calcsize(fmt)
            value = struct.unpack(
                fmt, self.buf[self.offset : self.offset + size]
            )
            if len(value) == 1:
                value = value[0]
            if isinstance(value, basestring) and sfmt[-1] == 's':
                value = value[: value.find(b'\x00')]
            if isinstance(sfmt, str) and sfmt[-1] == 's':
                try:
                    value = value.decode('utf-8')
                except UnicodeDecodeError:
                    value = hexdump(value)
            self[name] = routine(value)
            self.offset += size
        return self

    def encode(self):
        self._register_fields()
        for field in self.fields:
            name, fmt = field[:2]
            default = b'\x00' if len(field) <= 2 else field[2]
            fmt, routine = self._get_routine('encode', fmt)
            # special case: string
            if fmt == 'string':
                self.buf += routine(self[name])[0]
            else:
                size = struct.calcsize(fmt)
                if self[name] is None:
                    if not isinstance(default, basestring):
                        self.buf += struct.pack(fmt, default)
                    else:
                        self.buf += default * (size // len(default))
                else:
                    value = routine(self[name])
                    if not isinstance(value, (set, tuple, list)):
                        value = [value]
                    self.buf += struct.pack(fmt, *value)
        return self

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            if key in self._fields_names:
                return None
            raise


class ethmsg(msg):
    fields = (('dst', 'l2addr'), ('src', 'l2addr'), ('type', 'be16'))


class ip6msg(msg):
    fields = (
        ('version', 'uint8', 6 << 4),
        ('_flow0', 'uint8'),
        ('_flow1', 'uint8'),
        ('_flow2', 'uint8'),
        ('plen', 'uin16'),
        ('next_header', 'uint8'),
        ('hop_limit', 'uint8'),
        ('src', 'ip6addr'),
        ('dst', 'ip6addr'),
    )


class ip4msg(msg):
    fields = (
        ('verlen', 'uint8', 0x45),
        ('dsf', 'uint8'),
        ('len', 'be16'),
        ('id', 'be16'),
        ('flags', 'uint16'),
        ('ttl', 'uint8', 128),
        ('proto', 'uint8'),
        ('csum', 'be16'),
        ('src', 'ip4addr'),
        ('dst', 'ip4addr'),
    )


class udp4_pseudo_header(msg):
    fields = (
        ('src', 'ip4addr'),
        ('dst', 'ip4addr'),
        ('pad', 'uint8'),
        ('proto', 'uint8', 17),
        ('len', 'be16'),
    )


class udpmsg(msg):
    fields = (
        ('sport', 'be16'),
        ('dport', 'be16'),
        ('len', 'be16'),
        ('csum', 'be16'),
    )



#!/usr/bin/env python
import argparse
import json
import sys

from pyroute2.cli.console import Console
from pyroute2.cli.server import Server

try:
    from pyroute2.cli.auth.auth_keystone import OSAuthManager
except ImportError:
    OSAuthManager = None
try:
    from pyroute2.cli.auth.auth_radius import RadiusAuthManager
except ImportError:
    RadiusAuthManager = None


def run():
    argp = argparse.ArgumentParser()
    for spec in (
        ('-a', '[S] IP address to listen on'),
        ('-c', '[C] Command line to run'),
        ('-l', '[C,S] Log spec'),
        ('-m', 'Set mode (C,S)'),
        ('-p', '[S] Port to listen on'),
        ('-r', '[C] Load rc file'),
        ('-s', '[C,S] Load sources from a json file'),
        ('-x', '[S] Strict auth'),
    ):
        argp.add_argument(spec[0], help=spec[1])
    argp.add_argument('script', nargs='*', help='script to run')
    args = argp.parse_args()
    commands = []
    sources = None
    if args.s:
        with open(args.s, 'r') as f:
            sources = json.loads(f.read())

    if args.m in ('S', 'server'):
        if args.p:
            port = int(args.p)
        else:
            port = 8080

        auth_plugins = {}
        if OSAuthManager is not None:
            auth_plugins['keystone'] = OSAuthManager
        if RadiusAuthManager is not None:
            auth_plugins['radius:cleartext'] = RadiusAuthManager

        server = Server(
            address=args.a or 'localhost',
            port=port,
            log=args.l,
            sources=sources,
            auth_strict=args.x,
            auth_plugins=auth_plugins,
        )
        server.serve_forever()
        return 0
    else:
        console = Console(log=args.l, sources=sources)
        if args.r:
            console.loadrc(args.r)

        for script in args.script:
            console.loadrc(script)

        if args.c:
            commands.append(args.c)
            console.interact(readfunc=lambda x: commands.pop(0))
        elif not args.script:
            console.interact()
        return 1 if console.session.errors > 0 else 0


if __name__ == '__main__':
    rcode = run()
    sys.exit(rcode)



from pyroute2.common import map_namespace
from pyroute2.netlink import nlmsg_base

IFNAMSIZ = 16

IFF_UP = 0x1
IFF_BROADCAST = 0x2
IFF_DEBUG = 0x4
IFF_LOOPBACK = 0x8
IFF_POINTOPOINT = 0x10
IFF_STATICARP = 0x20
IFF_RUNNING = 0x40
IFF_NOARP = 0x80
IFF_PROMISC = 0x100
IFF_ALLMULTI = 0x200
IFF_OACTIVE = 0x400
IFF_SIMPLEX = 0x800
IFF_LINK0 = 0x1000
IFF_LINK1 = 0x2000
IFF_LINK2 = 0x4000
IFF_MULTICAST = 0x8000
(IFF_NAMES, IFF_VALUES) = map_namespace('IFF', globals())


class bsdmsg(nlmsg_base):
    __slots__ = ()
    header = (
        ('length', 'H'),
        ('version', 'B'),
        ('type', 'B'),
        ('hdrlen', 'H'),
    )


class if_msg(bsdmsg):
    __slots__ = ()
    fields = (
        ('ifm_index', 'H'),
        ('ifm_tableid', 'H'),
        ('ifm_pad1', 'B'),
        ('ifm_pad2', 'B'),
        ('ifm_addrs', 'i'),
        ('ifm_flags', 'i'),
        ('ifm_xflags', 'i'),
        ('ifi_type', 'B'),
        ('ifi_addrlen', 'B'),
        ('ifi_hdrlen', 'B'),
        ('ifi_link_state', 'B'),
        ('ifi_mtu', 'I'),
        ('ifi_metric', 'I'),
        ('ifi_rdomain', 'I'),
        ('ifi_baudrate', 'Q'),
        ('ifi_ipackets', 'Q'),
        ('ifi_ierrors', 'Q'),
        ('ifi_opackets', 'Q'),
        ('ifi_oerrors', 'Q'),
        ('ifi_collisions', 'Q'),
        ('ifi_ibytes', 'Q'),
        ('ifi_obytes', 'Q'),
        ('ifi_imcasts', 'Q'),
        ('ifi_omcasts', 'Q'),
        ('ifi_iqdrops', 'Q'),
        ('ifi_oqdrops', 'Q'),
        ('ifi_noproto', 'Q'),
        ('ifi_capabilities', 'I'),
        ('ifu_sec', 'Q'),
        ('ifu_usec', 'I'),
    )


class rt_msg_base(bsdmsg):
    __slots__ = ()
    fields = (
        ('rtm_index', 'H'),
        ('rtm_tableid', 'H'),
        ('rtm_priority', 'B'),
        ('rtm_mpls', 'B'),
        ('rtm_addrs', 'i'),
        ('rtm_flags', 'i'),
        ('rtm_fmask', 'i'),
        ('rtm_pid', 'I'),
        ('rtm_seq', 'i'),
        ('rtm_errno', 'i'),
        ('rtm_inits', 'I'),
        ('rmx_pksent', 'Q'),
        ('rmx_expire', 'q'),
        ('rmx_locks', 'I'),
        ('rmx_mtu', 'I'),
        ('rmx_refcnt', 'I'),
        ('rmx_hopcount', 'I'),
        ('rmx_recvpipe', 'I'),
        ('rmx_sendpipe', 'I'),
        ('rmx_ssthresh', 'I'),
        ('rmx_rtt', 'I'),
        ('rmx_rttvar', 'I'),
        ('rmx_pad', 'I'),
    )
    sockaddr_offset = 96

    ifa_slots = {
        0: ('DST', 'rt_slot_addr'),
        1: ('GATEWAY', 'rt_slot_addr'),
        2: ('NETMASK', 'rt_slot_addr'),
        3: ('GENMASK', 'hex'),
        4: ('IFP', 'rt_slot_ifp'),
        5: ('IFA', 'rt_slot_addr'),
        6: ('AUTHOR', 'hex'),
        7: ('BRD', 'rt_slot_addr'),
        8: ('SRC', 'rt_slot_addr'),
        9: ('SRCMASK', 'rt_slot_addr'),
        10: ('LABEL', 'hex'),
        11: ('BFD', 'hex'),
        12: ('DNS', 'hex'),
        13: ('STATIC', 'hex'),
        14: ('SEARCH', 'hex'),
    }


class ifa_msg_base(bsdmsg):
    __slots__ = ()
    fields = (
        ('ifam_index', 'H'),
        ('ifam_tableid', 'H'),
        ('ifam_pad1', 'B'),
        ('ifam_pad2', 'B'),
        ('rtm_addrs', 'i'),
        ('ifam_flags', 'i'),
        ('ifam_metric', 'i'),
    )
    sockaddr_offset = 24


class ifma_msg_base(bsdmsg):
    __slots__ = ()
    fields = (('rtm_addrs', 'i'), ('ifmam_flags', 'i'), ('ifmam_index', 'H'))
    sockaddr_offset = 16


class if_announcemsg(bsdmsg):
    __slots__ = ()
    fields = (
        ('ifan_index', 'H'),
        ('ifan_what', 'H'),
        ('ifan_name', '%is' % IFNAMSIZ),
    )

    def decode(self):
        bsdmsg.decode(self)
        self['ifan_name'] = self['ifan_name'].strip(b'\0').decode('ascii')



import struct
from socket import AF_INET, AF_INET6, AF_ROUTE, SOCK_RAW

from pyroute2 import config
from pyroute2.bsd.pf_route import (
    bsdmsg,
    if_announcemsg,
    if_msg,
    ifa_msg,
    ifma_msg,
    rt_msg,
)
from pyroute2.common import dqn2int
from pyroute2.netlink.rtnl import RTM_DELADDR as RTNL_DELADDR
from pyroute2.netlink.rtnl import RTM_DELLINK as RTNL_DELLINK
from pyroute2.netlink.rtnl import RTM_DELROUTE as RTNL_DELROUTE
from pyroute2.netlink.rtnl import RTM_NEWADDR as RTNL_NEWADDR
from pyroute2.netlink.rtnl import RTM_NEWLINK as RTNL_NEWLINK
from pyroute2.netlink.rtnl import RTM_NEWROUTE as RTNL_NEWROUTE
from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
from pyroute2.netlink.rtnl.rtmsg import rtmsg

if config.uname[0] == 'OpenBSD':
    from pyroute2.bsd.rtmsocket.openbsd import (
        RTM_ADD,
        RTM_NEWADDR,
        RTMSocketBase,
    )
else:
    from pyroute2.bsd.rtmsocket.freebsd import (
        RTM_ADD,
        RTM_NEWADDR,
        RTMSocketBase,
    )


def convert_rt_msg(msg):
    ret = rtmsg()
    ret['header']['type'] = (
        RTNL_NEWROUTE if msg['header']['type'] == RTM_ADD else RTNL_DELROUTE
    )
    ret['family'] = msg['DST']['header']['family']
    ret['attrs'] = []
    if 'address' in msg['DST']:
        ret['attrs'].append(['RTA_DST', msg['DST']['address']])
    if (
        'NETMASK' in msg
        and msg['NETMASK']['header']['family'] == ret['family']
    ):
        ret['dst_len'] = dqn2int(msg['NETMASK']['address'], ret['family'])
    if 'GATEWAY' in msg:
        if msg['GATEWAY']['header']['family'] not in (AF_INET, AF_INET6):
            # interface routes, table 255
            # discard for now
            return None
        ret['attrs'].append(['RTA_GATEWAY', msg['GATEWAY']['address']])
    if 'IFA' in msg:
        ret['attrs'].append(['RTA_SRC', msg['IFA']['address']])
    if 'IFP' in msg:
        ret['attrs'].append(['RTA_OIF', msg['IFP']['index']])
    elif msg['rtm_index'] != 0:
        ret['attrs'].append(['RTA_OIF', msg['rtm_index']])
    del ret['value']
    return ret


def convert_if_msg(msg):
    # discard this type for now
    return None


def convert_ifa_msg(msg):
    ret = ifaddrmsg()
    ret['header']['type'] = (
        RTNL_NEWADDR if msg['header']['type'] == RTM_NEWADDR else RTNL_DELADDR
    )
    ret['index'] = msg['IFP']['index']
    ret['family'] = msg['IFA']['header']['family']
    ret['prefixlen'] = dqn2int(msg['NETMASK']['address'], ret['family'])
    ret['attrs'] = [
        ['IFA_ADDRESS', msg['IFA']['address']],
        ['IFA_BROADCAST', msg['BRD']['address']],
        ['IFA_LABEL', msg['IFP']['ifname']],
    ]
    del ret['value']
    return ret


def convert_ifma_msg(msg):
    # ignore for now
    return None


def convert_if_announcemsg(msg):
    ret = ifinfmsg()
    ret['header']['type'] = RTNL_DELLINK if msg['ifan_what'] else RTNL_NEWLINK
    ret['index'] = msg['ifan_index']
    ret['attrs'] = [['IFLA_IFNAME', msg['ifan_name']]]
    del ret['value']
    return ret


def convert_bsdmsg(msg):
    # ignore unknown messages
    return None


convert = {
    rt_msg: convert_rt_msg,
    ifa_msg: convert_ifa_msg,
    if_msg: convert_if_msg,
    ifma_msg: convert_ifma_msg,
    if_announcemsg: convert_if_announcemsg,
    bsdmsg: convert_bsdmsg,
}


class RTMSocket(RTMSocketBase):
    def __init__(self, output='pf_route', target='localhost'):
        self.target = target
        self._sock = config.SocketBase(AF_ROUTE, SOCK_RAW)
        self._output = output

    def fileno(self):
        return self._sock.fileno()

    def get(self):
        msg = self._sock.recv(2048)
        _, _, msg_type = struct.unpack('HBB', msg[:4])
        msg_class = self.msg_map.get(msg_type, None)
        if msg_class is not None:
            msg = msg_class(msg)
            msg.decode()
            if self._output == 'netlink':
                # convert messages to the Netlink format
                msg = convert[type(msg)](msg)
                msg['header']['target'] = self.target
        return msg

    def close(self):
        self._sock.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


__all__ = [RTMSocket]



import re
import shlex

from pyroute2.cli import (
    t_comma,
    t_dict,
    t_end_of_dict,
    t_end_of_sentence,
    t_end_of_stream,
    t_pipe,
    t_stmt,
)
from pyroute2.common import basestring


class Token(object):
    def __init__(self, lex, expect=(), prohibit=(), leaf=False):
        self.lex = lex
        self.leaf = leaf
        self.kind = 0
        self.name = None
        self.argv = []
        self.kwarg = {}
        self.parse()
        if expect and self.kind not in expect:
            raise SyntaxError('expected %s, got %s' % (expect, self.kind))
        if prohibit and self.kind in prohibit:
            raise SyntaxError('unexpected %s' % (self.name,))

    def convert(self, arg):
        if re.match('^[0-9]+$', arg):
            return int(arg)
        else:
            return arg

    def parse(self):
        # triage
        first = self.lex.get_token()
        self.name = first

        ##
        # no token
        #
        if first == '':
            self.kind = t_end_of_stream

        ##
        # dict, e.g.
        #
        # resource spec, function arguments::
        #   {arg1, arg2}
        #   {key1 value1, key2 value2}
        #   {key {skey1 value}}
        #
        elif first == '{':
            arg_name = None
            while True:
                nt = Token(
                    self.lex, expect=(t_stmt, t_dict, t_comma, t_end_of_dict)
                )
                if arg_name is None:
                    if nt.kind == t_dict:
                        self.argv.append(nt.kwarg)
                    elif nt.kind == t_comma:
                        continue
                    elif nt.kind == t_stmt:
                        arg_name = nt.name
                else:
                    if nt.kind in (t_end_of_dict, t_comma):
                        self.argv.append(arg_name)
                    elif nt.kind == t_stmt:
                        self.kwarg[arg_name] = nt.name
                    elif nt.kind == t_dict:
                        self.kwarg[arg_name] = nt.kwarg
                    arg_name = None

                if nt.kind == t_end_of_dict:
                    self.kind = t_dict
                    self.name = '%s %s' % (self.argv, self.kwarg)
                    return

        ##
        # end of dict
        #
        elif first == '}':
            self.kind = t_end_of_dict

        ##
        # end of sentence
        #
        elif first == ';':
            self.kind = t_end_of_sentence

        ##
        # end of dict entry
        #
        elif first == ',':
            self.kind = t_comma

        ##
        # pipe
        #
        elif first == '|':
            self.kind = t_pipe

        elif first == '=':
            lookahead = self.lex.get_token()
            if lookahead == '>':
                self.name = '=>'
                self.kind = t_pipe
            else:
                self.lex.push_token(lookahead)
                self.kind = t_stmt

        ##
        # simple statement
        #
        # object name::
        #   name
        #
        # function call::
        #   func
        #   func {arg1, arg2}
        #   func {key1 value1, key2 value2}
        #
        else:
            self.name = self.convert(first)
            self.kind = t_stmt


class Sentence(object):
    def __init__(self, text, indent=0, master=None):
        self.offset = 0
        self.statements = []
        self.text = text
        self.lex = shlex.shlex(text)
        self.lex.wordchars += '.:/'
        self.lex.commenters = '#!'
        self.lex.debug = False
        self.indent = indent
        if master:
            self.chain = master.chain
        else:
            self.chain = []
            self.parse()

    def __iter__(self):
        for stmt in self.statements:
            yield stmt

    def parse(self):
        sentence = self
        while True:
            nt = Token(self.lex)
            if nt.kind == t_end_of_sentence:
                sentence = Sentence(None, self.indent, master=self)
            elif nt.kind == t_end_of_stream:
                return
            else:
                sentence.statements.append(nt)
            if sentence not in self.chain:
                self.chain.append(sentence)

    def __repr__(self):
        ret = '----\n'
        for s in self.statements:
            ret += '%i [%s] %s\n' % (self.indent, s.kind, s.name)
            ret += '\targv: %s\n' % (s.argv)
            ret += '\tkwarg: %s\n' % (s.kwarg)
        return ret


class Parser(object):
    def __init__(self, stream):
        self.stream = stream
        self.indent = None
        self.sentences = []
        self.parse()

    def parse(self):
        if hasattr(self.stream, 'readlines'):
            for text in self.stream.readlines():
                self.parse_string(text)
        elif isinstance(self.stream, basestring):
            self.parse_string(self.stream)
        else:
            raise ValueError('unsupported stream')
        self.parsed = True

    def parse_string(self, text):
        # 1. get indentation
        indent = re.match(r'^([ \t]*)', text).groups(0)[0]
        spaces = []
        # 2. sort it
        if indent:
            spaces = list(set(indent))
            if len(spaces) > 1:
                raise SyntaxError('mixed indentation')
            if self.indent is None:
                self.indent = spaces[0]
            if self.indent != spaces[0]:
                raise SyntaxError('mixed indentation')
        sentence = Sentence(text, len(indent))
        self.sentences.extend(sentence.chain)



import errno
import os
import threading

import mitogen.core
import mitogen.master

from pyroute2.iproute.linux import RTNL_API
from pyroute2.netlink.rtnl.iprsocket import MarshalRtnl

from .transport import RemoteSocket, Server, Transport


class Channel(object):
    def __init__(self, ch):
        self.ch = ch
        self._pfdr, self._pfdw = os.pipe()
        self.th = None
        self.closed = False
        self.lock = threading.RLock()
        self.shutdown_lock = threading.RLock()
        self.read = self._read_sync
        self.buf = ''

    def flush(self):
        pass

    def _read_sync(self, size):
        with self.lock:
            if self.buf:
                ret = self.buf[:size]
                self.buf = self.buf[size:]
                return ret
            ret = self.ch.get().unpickle()
            if len(ret) > size:
                self.buf = ret[size:]
            return ret[:size]

    def _read_async(self, size):
        with self.lock:
            return os.read(self._pfdr, size)

    def write(self, data):
        with self.lock:
            self.ch.send(data)
            return len(data)

    def start(self):
        with self.lock:
            if self.th is None:
                self.read = self._read_async
                self.th = threading.Thread(
                    target=self._monitor_thread,
                    name='Channel <%s> I/O' % self.ch,
                )
                self.th.start()

    def fileno(self):
        return self._pfdr

    def close(self):
        with self.shutdown_lock:
            if not self.closed:
                os.close(self._pfdw)
                os.close(self._pfdr)
                if self.th is not None:
                    self.th.join()
                self.closed = True
                if hasattr(self.ch, 'send'):
                    self.ch.send(None)

    def _monitor_thread(self):
        while True:
            msg = self.ch.get().unpickle()
            if msg is None:
                raise EOFError()
            os.write(self._pfdw, msg)


@mitogen.core.takes_router
def MitogenServer(ch_out, netns, target, router):
    ch_in = mitogen.core.Receiver(router)
    ch_out.send(ch_in.to_sender())

    trnsp_in = Transport(Channel(ch_in))
    trnsp_in.file_obj.start()
    trnsp_out = Transport(Channel(ch_out))

    return Server(trnsp_in, trnsp_out, netns, target)


class RemoteIPRoute(RTNL_API, RemoteSocket):
    def __init__(self, *argv, **kwarg):
        self._argv = tuple(argv)
        self._kwarg = dict(kwarg)
        if 'router' in kwarg:
            self._mitogen_broker = None
            self._mitogen_router = kwarg.pop('router')
        else:
            self._mitogen_broker = mitogen.master.Broker()
            self._mitogen_router = mitogen.master.Router(self._mitogen_broker)

        netns = kwarg.pop('netns', None)
        target = kwarg.pop('target', 'remote')
        try:
            if 'context' in kwarg:
                context = kwarg['context']
            else:
                protocol = kwarg.pop('protocol', 'local')
                context = getattr(self._mitogen_router, protocol)(
                    *argv, **kwarg
                )
            ch_in = mitogen.core.Receiver(
                self._mitogen_router, respondent=context
            )
            self._mitogen_call = context.call_async(
                MitogenServer,
                ch_out=ch_in.to_sender(),
                netns=netns,
                target=target,
            )
            ch_out = ch_in.get().unpickle()
            super(RemoteIPRoute, self).__init__(
                Transport(Channel(ch_in)), Transport(Channel(ch_out))
            )
        except Exception:
            if self._mitogen_broker is not None:
                self._mitogen_broker.shutdown()
                self._mitogen_broker.join()
            raise
        self.marshal = MarshalRtnl()
        self.target = target

    def clone(self):
        return type(self)(*self._argv, **self._kwarg)

    def close(self, code=errno.ECONNRESET):
        with self.shutdown_lock:
            if not self.closed:
                super(RemoteIPRoute, self).close(code=code)
                self.closed = True
                try:
                    self._mitogen_call.get()
                except mitogen.core.ChannelError:
                    pass
                if self._mitogen_broker is not None:
                    self._mitogen_broker.shutdown()
                    self._mitogen_broker.join()



##
#
# This module contains all the public symbols from the library.
#

##
#
# Version
#
try:
    from pyroute2.config.version import __version__
except ImportError:
    __version__ = 'unknown'


from pyroute2.conntrack import Conntrack, ConntrackEntry
from pyroute2.iproute import ChaoticIPRoute, IPBatch, IPRoute, RawIPRoute
from pyroute2.iproute.ipmock import IPRoute as IPMock
from pyroute2.iwutil import IW
from pyroute2.netlink.devlink import DevlinkSocket
from pyroute2.netlink.diag import DiagSocket
from pyroute2.netlink.event.acpi_event import AcpiEventSocket
from pyroute2.netlink.event.dquot import DQuotSocket
from pyroute2.netlink.exceptions import (
    ChaoticException,
    NetlinkDecodeError,
    NetlinkDumpInterrupted,
    NetlinkError,
)
from pyroute2.netlink.generic import GenericNetlinkSocket
from pyroute2.netlink.generic.l2tp import L2tp
from pyroute2.netlink.generic.mptcp import MPTCP
from pyroute2.netlink.generic.wireguard import WireGuard
from pyroute2.netlink.ipq import IPQSocket
from pyroute2.netlink.nfnetlink.nfctsocket import NFCTSocket
from pyroute2.netlink.nfnetlink.nftsocket import NFTSocket
from pyroute2.netlink.nl80211 import NL80211
from pyroute2.netlink.rtnl.iprsocket import IPRSocket
from pyroute2.netlink.taskstats import TaskStats
from pyroute2.netlink.uevent import UeventSocket

modules = [
    AcpiEventSocket,
    ChaoticException,
    ChaoticIPRoute,
    Conntrack,
    ConntrackEntry,
    DevlinkSocket,
    DiagSocket,
    DQuotSocket,
    IPBatch,
    IPMock,
    IPQSocket,
    IPRoute,
    IPRSocket,
    IW,
    GenericNetlinkSocket,
    L2tp,
    MPTCP,
    NetlinkError,
    NetlinkDecodeError,
    NetlinkDumpInterrupted,
    NFCTSocket,
    NFTSocket,
    NL80211,
    RawIPRoute,
    TaskStats,
    UeventSocket,
    WireGuard,
]

__all__ = []
__all__.extend(modules)



from pyroute2.netlink import nla

flags = {'reorder_hdr': 0x1, 'gvrp': 0x2, 'loose_binding': 0x4, 'mvrp': 0x8}


class vlan(nla):
    prefix = 'IFLA_'

    nla_map = (
        ('IFLA_VLAN_UNSPEC', 'none'),
        ('IFLA_VLAN_ID', 'uint16'),
        ('IFLA_VLAN_FLAGS', 'vlan_flags'),
        ('IFLA_VLAN_EGRESS_QOS', 'qos'),
        ('IFLA_VLAN_INGRESS_QOS', 'qos'),
        ('IFLA_VLAN_PROTOCOL', 'be16'),
    )

    class vlan_flags(nla):
        fields = (('flags', 'I'), ('mask', 'I'))

    class qos(nla):
        prefix = 'IFLA_'

        nla_map = (
            ('IFLA_VLAN_QOS_UNSPEC', 'none'),
            ('IFLA_VLAN_QOS_MAPPING', 'qos_mapping'),
        )

        class qos_mapping(nla):
            fields = (('from', 'I'), ('to', 'I'))



'''
'''
from enum import Enum

from pyroute2.netlink import genlmsg
from pyroute2.netlink.event import EventSocket
from pyroute2.netlink.nlsocket import Marshal


class ThermalGenlCmd(Enum):
    THERMAL_GENL_CMD_UNSPEC = 0
    THERMAL_GENL_CMD_TZ_GET_ID = 1
    THERMAL_GENL_CMD_TZ_GET_TRIP = 2
    THERMAL_GENL_CMD_TZ_GET_TEMP = 3
    THERMAL_GENL_CMD_TZ_GET_GOV = 4
    THERMAL_GENL_CMD_TZ_GET_MODE = 5
    THERMAL_GENL_CMD_CDEV_GET = 6


class ThermalGenlEvent(Enum):
    THERMAL_GENL_EVENT_UNSPEC = 0
    THERMAL_GENL_EVENT_TZ_CREATE = 1
    THERMAL_GENL_EVENT_TZ_DELETE = 2
    THERMAL_GENL_EVENT_TZ_DISABLE = 3
    THERMAL_GENL_EVENT_TZ_ENABLE = 4
    THERMAL_GENL_EVENT_TZ_TRIP_UP = 5
    THERMAL_GENL_EVENT_TZ_TRIP_DOWN = 6
    THERMAL_GENL_EVENT_TZ_TRIP_CHANGE = 7
    THERMAL_GENL_EVENT_TZ_TRIP_ADD = 8
    THERMAL_GENL_EVENT_TZ_TRIP_DELETE = 9
    THERMAL_GENL_EVENT_CDEV_ADD = 10
    THERMAL_GENL_EVENT_CDEV_DELETE = 11
    THERMAL_GENL_EVENT_CDEV_STATE_UPDATE = 12
    THERMAL_GENL_EVENT_TZ_GOV_CHANGE = 13
    THERMAL_GENL_EVENT_CPU_CAPABILITY_CHANGE = 14


class thermal_msg(genlmsg):
    nla_map = (
        ('THERMAL_GENL_ATTR_UNSPEC', 'none'),
        ('THERMAL_GENL_ATTR_TZ', 'nested'),
        ('THERMAL_GENL_ATTR_TZ_ID', 'uint32'),
        ('THERMAL_GENL_ATTR_TZ_TEMP', 'uint32'),
        ('THERMAL_GENL_ATTR_TZ_TRIP', 'nested'),
        ('THERMAL_GENL_ATTR_TZ_TRIP_ID', 'uint32'),
        ('THERMAL_GENL_ATTR_TZ_TRIP_TYPE', 'uint32'),
        ('THERMAL_GENL_ATTR_TZ_TRIP_TEMP', 'uint32'),
        ('THERMAL_GENL_ATTR_TZ_TRIP_HYST', 'uint32'),
        ('THERMAL_GENL_ATTR_TZ_MODE', 'uint32'),
        ('THERMAL_GENL_ATTR_TZ_NAME', 'asciiz'),
        ('THERMAL_GENL_ATTR_TZ_CDEV_WEIGHT', 'uint32'),
        ('THERMAL_GENL_ATTR_TZ_GOV', 'nested'),
        ('THERMAL_GENL_ATTR_TZ_GOV_NAME', 'asciiz'),
        ('THERMAL_GENL_ATTR_CDEV', 'nested'),
        ('THERMAL_GENL_ATTR_CDEV_ID', 'uint32'),
        ('THERMAL_GENL_ATTR_CDEV_CUR_STATE', 'uint32'),
        ('THERMAL_GENL_ATTR_CDEV_MAX_STATE', 'uint32'),
        ('THERMAL_GENL_ATTR_CDEV_NAME', 'asciiz'),
        ('THERMAL_GENL_ATTR_GOV_NAME', 'asciiz'),
        ('THERMAL_GENL_ATTR_CPU_CAPABILITY', 'nested'),
        ('THERMAL_GENL_ATTR_CPU_CAPABILITY_ID', 'uint32'),
        ('THERMAL_GENL_ATTR_CPU_CAPABILITY_PERFORMANCE', 'uint32'),
        ('THERMAL_GENL_ATTR_CPU_CAPABILITY_EFFICIENCY', 'uint32'),
    )


class MarshalThermalEvent(Marshal):
    msg_map = {x.value: thermal_msg for x in ThermalGenlEvent}


class ThermalEventSocket(EventSocket):
    marshal_class = MarshalThermalEvent
    genl_family = 'thermal'



from struct import pack, unpack

from pyroute2.netlink import nla

TCF_EM_OPND_EQ = 0
TCF_EM_OPND_GT = 1
TCF_EM_OPND_LT = 2

OPERANDS_DICT = {
    TCF_EM_OPND_EQ: ('eq', '='),
    TCF_EM_OPND_GT: ('gt', '>'),
    TCF_EM_OPND_LT: ('lt', '<'),
}

# meta types
TCF_META_TYPE_VAR = 0
TCF_META_TYPE_INT = 1

TCF_META_ID_MASK = 0x7FF
TCF_META_TYPE_MASK = 0xF << 12

# see tc_em_meta.h
META_ID = {
    'value': 0,
    'random': 1,
    'loadavg_0': 2,
    'loadavg_1': 3,
    'loadavg_2': 4,
    'dev': 5,
    'priority': 6,
    'protocol': 7,
    'pkttype': 8,
    'pktlen': 9,
    'datalen': 10,
    'maclen': 11,
    'nfmark': 12,
    'tcindex': 13,
    'rtclassid': 14,
    'rtiif': 15,
    'sk_family': 16,
    'sk_state': 17,
    'sk_reuse': 18,
    'sk_bound_if': 19,
    'sk_refcnt': 20,
    'sk_shutdown': 21,
    'sk_proto': 22,
    'sk_type': 23,
    'sk_rcvbuf': 24,
    'sk_rmem_alloc': 25,
    'sk_wmem_alloc': 26,
    'sk_omem_alloc': 27,
    'sk_wmem_queued': 28,
    'sk_rcv_qlen': 29,
    'sk_snd_qlen': 30,
    'sk_err_qlen': 31,
    'sk_forward_allocs': 32,
    'sk_sndbuf': 33,
    'sk_allocs': 34,
    'sk_route_caps': 35,
    'sk_hash': 36,
    'sk_lingertime': 37,
    'sk_ack_backlog': 38,
    'sk_max_ack_backlog': 39,
    'sk_prio': 40,
    'sk_rcvlowat': 41,
    'sk_rcvtimeo': 42,
    'sk_sndtimeo': 43,
    'sk_sendmsg_off': 44,
    'sk_write_pending': 45,
    'vlan_tag': 46,
    'rxhash': 47,
}

strings_meta = ('dev', 'sk_bound_if')


class data(nla):
    nla_map = (
        ('TCA_EM_META_UNSPEC', 'none'),
        ('TCA_EM_META_HDR', 'tca_em_meta_header_parse'),
        ('TCA_EM_META_LVALUE', 'uint32'),
        ('TCA_EM_META_RVALUE', 'hex'),
    )

    def decode(self):
        self.header = None
        self.length = 24
        nla.decode(self)

        # Patch to have a better view in nldecap
        attrs = dict(self['attrs'])
        rvalue = attrs.get('TCA_EM_META_RVALUE')
        meta_hdr = attrs.get('TCA_EM_META_HDR')
        meta_id = meta_hdr['id']
        rvalue = bytearray.fromhex(rvalue.replace(':', ''))
        if meta_id == 'TCF_META_TYPE_VAR':
            rvalue.decode('utf-8')
        if meta_id == 'TCF_META_TYPE_INT':
            rvalue = unpack('<I', rvalue)[0]
        self['attrs'][2] = ('TCA_EM_META_RVALUE', rvalue)

    def encode(self):
        if 'object' not in self:
            raise ValueError('An object definition must be given!')

        if 'value' not in self:
            raise ValueError('A value must be given!')

        # The value can either be a string or an int depending
        # on the selected meta kind.
        # It is really important to distinct them otherwise
        # the kernel won't be nice with us...
        value = self['value']
        kind = self['object'].get('kind')
        if not kind:
            raise ValueError('Not meta kind specified!')
        else:
            if kind in strings_meta:
                if not isinstance(value, str):
                    raise ValueError(
                        '{} kinds have to use string value!'.format(
                            ' and '.join(strings_meta)
                        )
                    )
                else:
                    value = value.encode('utf-8')
            else:
                if not isinstance(value, int):
                    raise ValueError(
                        'Invalid value specified, it must ' 'be an integer'
                    )
                else:
                    value = pack('<I', value)

        self['attrs'].append(['TCA_EM_META_HDR', self['object']])
        self['attrs'].append(['TCA_EM_META_LVALUE', self.get('mask', 0)])
        self['attrs'].append(['TCA_EM_META_RVALUE', value])
        nla.encode(self)

        # Patch NLA structure
        self['header']['length'] -= 4
        self.data = self.data[4:]

    class tca_em_meta_header_parse(nla):
        fields = (
            ('kind', 'H'),
            ('shift', 'B'),
            ('opnd', 'B'),
            ('id', 'H'),
            ('pad', 'H'),
        )

        def decode(self):
            nla.decode(self)

            self['id'] = (self['kind'] & TCF_META_TYPE_MASK) >> 12
            if self['id'] == TCF_META_TYPE_VAR:
                self['id'] = 'TCF_META_TYPE_VAR'
            elif self['id'] == TCF_META_TYPE_INT:
                self['id'] = 'TCF_META_TYPE_INT'
            else:
                pass

            self['kind'] &= TCF_META_ID_MASK
            for k, v in META_ID.items():
                if self['kind'] == v:
                    self['kind'] = 'TCF_META_ID_{}'.format(k.upper())

            fmt = 'TCF_EM_OPND_{}'.format(
                OPERANDS_DICT[self['opnd']][0].upper()
            )
            self['opnd'] = fmt
            del self['pad']

        def encode(self):
            if not isinstance(self['kind'], str):
                raise ValueError("kind' keywords must be set!")

            kind = self['kind'].lower()
            if kind in strings_meta:
                self['id'] = TCF_META_TYPE_VAR
            else:
                self['id'] = TCF_META_TYPE_INT
            self['id'] <<= 12

            for k, v in META_ID.items():
                if kind == k:
                    self['kind'] = self['id'] | v
                    break

            if isinstance(self['opnd'], str):
                for k, v in OPERANDS_DICT.items():
                    if self['opnd'].lower() in v:
                        self['opnd'] = k
                        break

            # Perform sanity checks on 'shift' value
            if isinstance(self['shift'], str):
                # If it fails, it will raise a ValueError
                # which is what we want
                self['shift'] = int(self['shift'])
            if not 0 <= self['shift'] <= 255:
                raise ValueError(
                    "'shift' value must be between" "0 and 255 included!"
                )
            nla.encode(self)



try:
    from multiprocessing import shared_memory
except ImportError:
    shared_memory = None


class Page:
    '''
    Memory page.
    '''

    def __init__(self, view, offset):
        self.view = view
        self.offset = offset
        self.is_free = True

    def use(self):
        self.is_free = False

    def free(self):
        self.is_free = True

    def close(self):
        self.view.release()


class Buffer:
    '''
    Manage the buffer memory to receive raw netlink data.
    '''

    def __init__(self, mode='internal', size=10485760, page_size=32768):
        self.mode = mode
        self.size = size
        self.page_size = page_size
        if self.mode == 'internal':
            self.mem = None
            self.buf = bytearray(self.size)
        elif self.mode == 'shared':
            if shared_memory is None:
                raise ModuleNotFoundError('shared memory buffer not supported')
            self.mem = shared_memory.SharedMemory(create=True, size=self.size)
            self.buf = self.mem.buf
        self.view = memoryview(self.buf)
        self.directory = {}
        for index in range(size // page_size):
            offset = index * page_size
            self.directory[index] = Page(
                self.view[offset : offset + self.page_size], offset
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def get_free_page(self):
        for index, page in self.directory.items():
            if page.is_free:
                page.use()
                return page
        raise MemoryError('no free memory pages available')

    def close(self):
        for page in self.directory.values():
            page.close()
        self.view.release()
        if self.mode == 'shared':
            self.mem.close()
            self.mem.unlink()

    def __getitem__(self, key):
        return self.directory[key]



'''
cake
++++

Usage:

    # Imports
    from pyroute2 import IPRoute


    # Add cake with 2048kbit of bandwidth capacity
    with IPRoute() as ipr:
        # Get interface index
        index = ipr.link_lookup(ifname='lo')
        ipr.tc('add', kind='cake', index=index, bandwidth='2048kbit')

    # Same with 15mbit of bandwidth capacity
    with IPRoute() as ipr:
        # Get interface index
        index = ipr.link_lookup(ifname='lo')
        ipr.tc('add', kind='cake', index=index, bandwidth='15mbit')

    # If you don't know the bandwidth capacity, use autorate
    with IPRoute() as ipr:
        # Get interface index
        index = ipr.link_lookup(ifname='lo')
        ipr.tc('add', kind='cake', index=index, bandwidth='unlimited',
               autorate=True)

    # If you want to tune ATM properties use:
    # atm_mode=False # For no ATM tuning
    # atm_mode=True # For ADSL tuning
    # atm_mode='ptm' # For VDSL2 tuning
    with IPRoute() as ipr:
        # Get interface index
        index = ipr.link_lookup(ifname='lo')
        ipr.tc('add', kind='cake', index=index, bandwidth='unlimited',
               autorate=True, atm_mode=True)

    # Complex example which has no-sense
    with IPRoute() as ipr:
        # Get interface index
        index = ipr.link_lookup(ifname='lo')
        ipr.tc('add', kind='cake', index=index, bandwidth='unlimited',
               autorate=True, nat=True, rtt='interplanetary', target=10000,
               flow_mode='dsthost', diffserv_mode='precedence', fwmark=0x1337)

NOTES:
    Here is the list of all supported options with their values:
    - ack_filter: False, True or 'aggressive' (False by default)
    - atm_mode: False, True or 'ptm' (False by default)
    - autorate: False or True (False by default)
    - bandwidth: any integer, 'N[kbit|mbit|gbit]' or 'unlimited'
    - diffserv_mode: 'diffserv3', 'diffserv4', 'diffserv8',
        'besteffort', 'precedence' ('diffserv3' by default)
    - ingress: False or True (False by default)
    - overhead: any integer between -64 and 256 inclusive (0 by default)
    - flow_mode: 'flowblind', 'srchost', 'dsthost', 'hosts', 'flows',
        'dual-srchost', 'dual-dsthost', 'triple-isolate'
        ('triple-isolate' by default)
    - fwmark: any integer (0 by default)
    - memlimit: any integer (by default, calculated based on the bandwidth
        and RTT settings)
    - mpu: any integer between 0 and 256 inclusive (0 by default)
    - nat: False or True (False by default)
    - raw: False or True (True by default)
    - rtt: any integer or 'datacentre', 'lan', 'metro', 'regional',
        'internet', 'oceanic', 'satellite', 'interplanetary'
        ('internet' by default)
    - split_gso: False or True (True by default)
    - target: any integer (5000 by default)
    - wash: False or True (False by default)
'''


from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import TC_H_ROOT

# Defines from pkt_sched.h
CAKE_FLOW_NONE = 0
CAKE_FLOW_SRC_IP = 1
CAKE_FLOW_DST_IP = 2
CAKE_FLOW_HOSTS = 3
CAKE_FLOW_FLOWS = 4
CAKE_FLOW_DUAL_SRC = 5
CAKE_FLOW_DUAL_DST = 6
CAKE_FLOW_TRIPLE = 7

CAKE_DIFFSERV_DIFFSERV3 = 0
CAKE_DIFFSERV_DIFFSERV4 = 1
CAKE_DIFFSERV_DIFFSERV8 = 2
CAKE_DIFFSERV_BESTEFFORT = 3
CAKE_DIFFSERV_PRECEDENCE = 4

CAKE_ACK_NONE = 0
CAKE_ACK_FILTER = 1
CAKE_ACK_AGGRESSIVE = 2

CAKE_ATM_NONE = 0
CAKE_ATM_ATM = 1
CAKE_ATM_PTM = 2

TCA_CAKE_MAX_TINS = 8


def fix_msg(msg, kwarg):
    if 'parent' not in kwarg:
        msg['parent'] = TC_H_ROOT


def convert_bandwidth(value):
    types = [('kbit', 1000), ('mbit', 1000000), ('gbit', 1000000000)]

    if 'unlimited' == value:
        return 0

    try:
        # Value is passed as an int
        x = int(value)
        return x >> 3
    except ValueError:
        value = value.lower()
        for t, mul in types:
            if len(value.split(t)) == 2:
                x = int(value.split(t)[0]) * mul
                return x >> 3

    raise ValueError(
        'Invalid bandwidth value. Specify either an integer, '
        '"unlimited" or a value with "kbit", "mbit" or '
        '"gbit" appended'
    )


def convert_rtt(value):
    types = {
        'datacentre': 100,
        'lan': 1000,
        'metro': 10000,
        'regional': 30000,
        'internet': 100000,
        'oceanic': 300000,
        'satellite': 1000000,
        'interplanetary': 3600000000,
    }

    try:
        # Value is passed as an int
        x = int(value)
        return x
    except ValueError:
        rtt = types.get(value.lower())
        if rtt is not None:
            return rtt
        raise ValueError(
            'Invalid rtt value. Specify either an integer (us), '
            'or datacentre, lan, metro, regional, internet, '
            'oceanic or interplanetary.'
        )


def convert_atm(value):
    if isinstance(value, bool):
        if not value:
            return CAKE_ATM_NONE
        else:
            return CAKE_ATM_ATM
    else:
        if value == 'ptm':
            return CAKE_ATM_PTM

    raise ValueError('Invalid ATM value!')


def convert_flowmode(value):
    modes = {
        'flowblind': CAKE_FLOW_NONE,
        'srchost': CAKE_FLOW_SRC_IP,
        'dsthost': CAKE_FLOW_DST_IP,
        'hosts': CAKE_FLOW_HOSTS,
        'flows': CAKE_FLOW_FLOWS,
        'dual-srchost': CAKE_FLOW_DUAL_SRC,
        'dual-dsthost': CAKE_FLOW_DUAL_DST,
        'triple-isolate': CAKE_FLOW_TRIPLE,
    }

    res = modes.get(value.lower())
    if res:
        return res
    raise ValueError(
        'Invalid flow mode specified! See tc-cake man '
        'page for valid values.'
    )


def convert_diffserv(value):
    modes = {
        'diffserv3': CAKE_DIFFSERV_DIFFSERV3,
        'diffserv4': CAKE_DIFFSERV_DIFFSERV4,
        'diffserv8': CAKE_DIFFSERV_DIFFSERV8,
        'besteffort': CAKE_DIFFSERV_BESTEFFORT,
        'precedence': CAKE_DIFFSERV_PRECEDENCE,
    }

    res = modes.get(value.lower())
    if res is not None:
        return res
    raise ValueError(
        'Invalid diffserv mode specified! See tc-cake man '
        'page for valid values.'
    )


def convert_ackfilter(value):
    if isinstance(value, bool):
        if not value:
            return CAKE_ACK_NONE
        else:
            return CAKE_ACK_FILTER
    else:
        if value == 'aggressive':
            return CAKE_ACK_AGGRESSIVE

    raise ValueError('Invalid ACK filter!')


def check_range(name, value, start, end):
    if not isinstance(value, int):
        raise ValueError('{} value must be an integer'.format(name))

    if not start <= value <= end:
        raise ValueError(
            '{0} value must be between {1} and {2} '
            'inclusive.'.format(name, start, end)
        )


def get_parameters(kwarg):
    ret = {'attrs': []}
    attrs_map = (
        ('ack_filter', 'TCA_CAKE_ACK_FILTER'),
        ('atm_mode', 'TCA_CAKE_ATM'),
        ('autorate', 'TCA_CAKE_AUTORATE'),
        ('bandwidth', 'TCA_CAKE_BASE_RATE64'),
        ('diffserv_mode', 'TCA_CAKE_DIFFSERV_MODE'),
        ('ingress', 'TCA_CAKE_INGRESS'),
        ('overhead', 'TCA_CAKE_OVERHEAD'),
        ('flow_mode', 'TCA_CAKE_FLOW_MODE'),
        ('fwmark', 'TCA_CAKE_FWMARK'),
        ('memlimit', 'TCA_CAKE_MEMORY'),
        ('mpu', 'TCA_CAKE_MPU'),
        ('nat', 'TCA_CAKE_NAT'),
        ('raw', 'TCA_CAKE_RAW'),
        ('rtt', 'TCA_CAKE_RTT'),
        ('split_gso', 'TCA_CAKE_SPLIT_GSO'),
        ('target', 'TCA_CAKE_TARGET'),
        ('wash', 'TCA_CAKE_WASH'),
    )

    for k, v in attrs_map:
        r = kwarg.get(k, None)
        if r is not None:
            if k == 'bandwidth':
                r = convert_bandwidth(r)
            elif k == 'rtt':
                r = convert_rtt(r)
            elif k == 'atm_mode':
                r = convert_atm(r)
            elif k == 'flow_mode':
                r = convert_flowmode(r)
            elif k == 'diffserv_mode':
                r = convert_diffserv(r)
            elif k == 'ack_filter':
                r = convert_ackfilter(r)
            elif k == 'mpu':
                check_range(k, r, 0, 256)
            elif k == 'overhead':
                check_range(k, r, -64, 256)
            ret['attrs'].append([v, r])

    return ret


class options(nla):
    nla_map = (
        ('TCA_CAKE_UNSPEC', 'none'),
        ('TCA_CAKE_PAD', 'uint64'),
        ('TCA_CAKE_BASE_RATE64', 'uint64'),
        ('TCA_CAKE_DIFFSERV_MODE', 'uint32'),
        ('TCA_CAKE_ATM', 'uint32'),
        ('TCA_CAKE_FLOW_MODE', 'uint32'),
        ('TCA_CAKE_OVERHEAD', 'int32'),
        ('TCA_CAKE_RTT', 'uint32'),
        ('TCA_CAKE_TARGET', 'uint32'),
        ('TCA_CAKE_AUTORATE', 'uint32'),
        ('TCA_CAKE_MEMORY', 'uint32'),
        ('TCA_CAKE_NAT', 'uint32'),
        ('TCA_CAKE_RAW', 'uint32'),
        ('TCA_CAKE_WASH', 'uint32'),
        ('TCA_CAKE_MPU', 'uint32'),
        ('TCA_CAKE_INGRESS', 'uint32'),
        ('TCA_CAKE_ACK_FILTER', 'uint32'),
        ('TCA_CAKE_SPLIT_GSO', 'uint32'),
        ('TCA_CAKE_FWMARK', 'uint32'),
    )

    def encode(self):
        # Set default Auto-Rate value
        if not self.get_attr('TCA_CAKE_AUTORATE'):
            self['attrs'].append(['TCA_CAKE_AUTORATE', 0])
        nla.encode(self)


class stats2(nla):
    nla_map = (
        ('TCA_STATS_UNSPEC', 'none'),
        ('TCA_STATS_BASIC', 'basic'),
        ('TCA_STATS_RATE_EST', 'rate_est'),
        ('TCA_STATS_QUEUE', 'queue'),
        ('TCA_STATS_APP', 'stats_app'),
    )

    class basic(nla):
        fields = (('bytes', 'Q'), ('packets', 'I'))

    class rate_est(nla):
        fields = (('bps', 'I'), ('pps', 'I'))

    class queue(nla):
        fields = (
            ('qlen', 'I'),
            ('backlog', 'I'),
            ('drops', 'I'),
            ('requeues', 'I'),
            ('overlimits', 'I'),
        )

    class stats_app(nla):
        nla_map = (
            ('__TCA_CAKE_STATS_INVALID', 'none'),
            ('TCA_CAKE_STATS_PAD', 'hex'),
            ('TCA_CAKE_STATS_CAPACITY_ESTIMATE64', 'uint64'),
            ('TCA_CAKE_STATS_MEMORY_LIMIT', 'uint32'),
            ('TCA_CAKE_STATS_MEMORY_USED', 'uint32'),
            ('TCA_CAKE_STATS_AVG_NETOFF', 'uint32'),
            ('TCA_CAKE_STATS_MAX_NETLEN', 'uint32'),
            ('TCA_CAKE_STATS_MAX_ADJLEN', 'uint32'),
            ('TCA_CAKE_STATS_MIN_NETLEN', 'uint32'),
            ('TCA_CAKE_STATS_MIN_ADJLEN', 'uint32'),
            ('TCA_CAKE_STATS_TIN_STATS', 'tca_parse_tins'),
            ('TCA_CAKE_STATS_DEFICIT', 'uint32'),
            ('TCA_CAKE_STATS_COBALT_COUNT', 'uint32'),
            ('TCA_CAKE_STATS_DROPPING', 'uint32'),
            ('TCA_CAKE_STATS_DROP_NEXT_US', 'uint32'),
            ('TCA_CAKE_STATS_P_DROP', 'uint32'),
            ('TCA_CAKE_STATS_BLUE_TIMER_US', 'uint32'),
        )

        class tca_parse_tins(nla):
            nla_map = tuple(
                [
                    ('TCA_CAKE_TIN_STATS_%i' % x, 'tca_parse_tin_stats')
                    for x in range(TCA_CAKE_MAX_TINS)
                ]
            )

            class tca_parse_tin_stats(nla):
                nla_map = (
                    ('__TCA_CAKE_TIN_STATS_INVALID', 'none'),
                    ('TCA_CAKE_TIN_STATS_PAD', 'hex'),
                    ('TCA_CAKE_TIN_STATS_SENT_PACKETS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_SENT_BYTES64', 'uint64'),
                    ('TCA_CAKE_TIN_STATS_DROPPED_PACKETS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_DROPPED_BYTES64', 'uint64'),
                    ('TCA_CAKE_TIN_STATS_ACKS_DROPPED_PACKETS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_ACKS_DROPPED_BYTES64', 'uint64'),
                    ('TCA_CAKE_TIN_STATS_ECN_MARKED_PACKETS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_ECN_MARKED_BYTES64', 'uint64'),
                    ('TCA_CAKE_TIN_STATS_BACKLOG_PACKETS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_BACKLOG_BYTES', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_THRESHOLD_RATE64', 'uint64'),
                    ('TCA_CAKE_TIN_STATS_TARGET_US', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_INTERVAL_US', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_WAY_INDIRECT_HITS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_WAY_MISSES', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_WAY_COLLISIONS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_PEAK_DELAY_US', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_AVG_DELAY_US', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_BASE_DELAY_US', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_SPARSE_FLOWS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_BULK_FLOWS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_UNRESPONSIVE_FLOWS', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_MAX_SKBLEN', 'uint32'),
                    ('TCA_CAKE_TIN_STATS_FLOW_QUANTUM', 'uint32'),
                )



import struct
import sys

##
#
# Logging setup
#
# See the history:
#  * https://github.com/svinota/pyroute2/issues/246
#  * https://github.com/svinota/pyroute2/issues/255
#  * https://github.com/svinota/pyroute2/issues/270
#  * https://github.com/svinota/pyroute2/issues/573
#  * https://github.com/svinota/pyroute2/issues/601
#
from pyroute2.config import log

##
#
# Windows platform specific: socket module monkey patching
#
# To use the library on Windows, run::
#   pip install win-inet-pton
#
if sys.platform.startswith('win'):  # noqa: E402
    import win_inet_pton  # noqa: F401


def init():
    try:
        # probe, if the bytearray can be used in struct.unpack_from()
        struct.unpack_from('I', bytearray((1, 0, 0, 0)), 0)
    except Exception:
        if sys.version_info[0] < 3:
            # monkeypatch for old Python versions
            log.warning('patching struct.unpack_from()')

            def wrapped(fmt, buf, offset=0):
                return struct._u_f_orig(fmt, str(buf), offset)

            struct._u_f_orig = struct.unpack_from
            struct.unpack_from = wrapped
        else:
            raise



from pyroute2.netlink import nla
from pyroute2.netlink.rtnl import TC_H_ROOT

parent = TC_H_ROOT


class options(nla):
    fields = (('bands', 'i'), ('priomap', '16B'))


def get_parameters(kwarg):
    return kwarg



'''
Backends
--------

NDB stores all the records in an SQL database. By default it uses
the SQLite3 module, which is a part of the Python stdlib, so no
extra packages are required::

    # SQLite3 -- simple in-memory DB
    ndb = NDB()

    # SQLite3 -- same as above with explicit arguments
    ndb = NDB(db_provider='sqlite3', db_spec=':memory:')

    # SQLite3 -- file DB
    ndb = NDB(db_provider='sqlite3', db_spec='test.db')

It is also possible to use a PostgreSQL database via psycopg2
module::

    # PostgreSQL -- local DB
    ndb = NDB(db_provider='psycopg2',
              db_spec={'dbname': 'test'})

    # PostgreSQL -- remote DB
    ndb = NDB(db_provider='psycopg2',
              db_spec={'dbname': 'test',
                       'host': 'db1.example.com'})

Database backup
---------------

Built-in database backup is implemented now only for SQLite3 backend.
For the PostgresSQL backend you have to use external utilities like
`pg_dump`::

    # create an NDB instance
    ndb = NDB()  # the defaults: db_provider='sqlite3', db_spec=':memory:'
    ...
    # dump the DB to a file
    ndb.backup('backup.db')

SQL schema
----------

By default NDB deletes the data from the DB upon exit. In order to preserve
the data, use `NDB(db_cleanup=False, ...)`

Here is an example schema (may be changed with releases)::

                List of relations
     Schema |       Name       | Type  | Owner
    --------+------------------+-------+-------
     public | addresses        | table | root
     public | af_bridge_fdb    | table | root
     public | af_bridge_ifs    | table | root
     public | af_bridge_vlans  | table | root
     public | enc_mpls         | table | root
     public | ifinfo_bond      | table | root
     public | ifinfo_bridge    | table | root
     public | ifinfo_gre       | table | root
     public | ifinfo_gretap    | table | root
     public | ifinfo_ip6gre    | table | root
     public | ifinfo_ip6gretap | table | root
     public | ifinfo_ip6tnl    | table | root
     public | ifinfo_ipip      | table | root
     public | ifinfo_ipvlan    | table | root
     public | ifinfo_macvlan   | table | root
     public | ifinfo_macvtap   | table | root
     public | ifinfo_sit       | table | root
     public | ifinfo_tun       | table | root
     public | ifinfo_vlan      | table | root
     public | ifinfo_vrf       | table | root
     public | ifinfo_vti       | table | root
     public | ifinfo_vti6      | table | root
     public | ifinfo_vxlan     | table | root
     public | interfaces       | table | root
     public | metrics          | table | root
     public | neighbours       | table | root
     public | netns            | table | root
     public | nh               | table | root
     public | p2p              | table | root
     public | routes           | table | root
     public | rules            | table | root
     public | sources          | table | root
     public | sources_options  | table | root
    (33 rows)

    rtnl=# select f_index, f_ifla_ifname from interfaces;
     f_index | f_ifla_ifname
    ---------+---------------
           1 | lo
           2 | eth0
          28 | ip_vti0
          31 | ip6tnl0
          32 | ip6_vti0
       36445 | br0
       11434 | dummy0
           3 | eth1
    (8 rows)

    rtnl=# select f_index, f_ifla_br_stp_state from ifinfo_bridge;
     f_index | f_ifla_br_stp_state
    ---------+---------------------
       36445 |                   0
    (1 row)

Database upgrade
----------------

There is no DB schema upgrade from release to release. All the
data stored in the DB is being fetched from the OS in the runtime,
thus no persistence required.

If you're using a PostgreSQL DB or a file based SQLite, simply drop
all the tables from the DB, and NDB will create them from scratch
on startup.
'''
import enum
import json
import random
import sqlite3
import sys
import time
import traceback
from collections import OrderedDict
from functools import partial

from pyroute2 import config
from pyroute2.common import basestring, uuid32

#
from .objects import address, interface, neighbour, netns, route, rule

try:
    import psycopg2
except ImportError:
    psycopg2 = None

#
# the order is important
#
plugins = [interface, address, neighbour, route, netns, rule]

MAX_ATTEMPTS = 5


class DBProvider(enum.Enum):
    sqlite3 = 'sqlite3'
    psycopg2 = 'psycopg2'

    def __eq__(self, r):
        return str(self) == r


def publish(f):
    if isinstance(f, str):

        def decorate(m):
            m.publish = f
            return m

        return decorate

    f.publish = True
    return f


class DBDict(dict):
    def __init__(self, schema, table):
        self.schema = schema
        self.table = table

    @publish('get')
    def __getitem__(self, key):
        for (record,) in self.schema.fetch(
            f'''
            SELECT f_value FROM {self.table}
            WHERE f_key = {self.schema.plch}
            ''',
            (key,),
        ):
            return json.loads(record)
        raise KeyError(f'key {key} not found')

    @publish('set')
    def __setitem__(self, key, value):
        del self[key]
        self.schema.execute(
            f'''
            INSERT INTO {self.table}
            VALUES ({self.schema.plch}, {self.schema.plch})
            ''',
            (key, json.dumps(value)),
        )

    @publish('del')
    def __delitem__(self, key):
        self.schema.execute(
            f'''
            DELETE FROM {self.table}
            WHERE f_key = {self.schema.plch}
            ''',
            (key,),
        )

    @publish
    def keys(self):
        for (key,) in self.schema.fetch(f'SELECT f_key FROM {self.table}'):
            yield key

    @publish
    def items(self):
        for key, value in self.schema.fetch(
            f'SELECT f_key, f_value FROM {self.table}'
        ):
            yield key, json.loads(value)

    @publish
    def values(self):
        for (value,) in self.schema.fetch(f'SELECT f_value FROM {self.table}'):
            yield json.loads(value)


class DBSchema:
    connection = None
    event_map = None
    key_defaults = None
    snapshots = None  # <table_name>: <obj_weakref>

    spec = OrderedDict()
    classes = {}
    #
    # OBS: field names MUST go in the same order as in the spec,
    # that's for the load_netlink() to work correctly -- it uses
    # one loop to fetch both index and row values
    #
    indices = {}
    foreign_keys = {}

    def __init__(self, config, sources, event_map, log_channel):
        global plugins
        self.sources = sources
        self.config = DBDict(self, 'config')
        self.stats = {}
        self.connection = None
        self.cursor = None
        self.log = log_channel
        self.snapshots = {}
        self.key_defaults = {}
        self.event_map = {}
        # cache locally these variables so they will not be
        # loaded from SQL for every incoming message; this
        # means also that these variables can not be changed
        # in runtime
        self.rtnl_log = config['rtnl_debug']
        self.provider = config['provider']
        #
        for plugin in plugins:
            #
            # 1. spec
            #
            for name, spec in plugin.init['specs']:
                self.spec[name] = spec.as_dict()
                self.indices[name] = spec.index
                self.foreign_keys[name] = spec.foreign_keys
            #
            # 2. classes
            #
            for name, cls in plugin.init['classes']:
                self.classes[name] = cls
        #
        self.initdb(config)
        #
        for plugin in plugins:
            #
            emap = plugin.init['event_map']
            #
            for etype, ehndl in emap.items():
                handlers = []
                for h in ehndl:
                    if isinstance(h, basestring):
                        handlers.append(partial(self.load_netlink, h))
                    else:
                        handlers.append(partial(h, self))
                self.event_map[etype] = handlers

        self.gctime = self.ctime = time.time()

    def initdb(self, config):
        if self.connection is not None:
            self.close()
        if config['provider'] == DBProvider.sqlite3:
            self.connection = sqlite3.connect(config['spec'])
            self.plch = '?'
            self.connection.execute('PRAGMA foreign_keys = ON')
        elif config['provider'] == DBProvider.psycopg2:
            self.connection = psycopg2.connect(**config['spec'])
            self.plch = '%s'
        else:
            raise TypeError('DB provider not supported')
        self.cursor = self.connection.cursor()
        #
        # compile request lines
        #
        self.compiled = {}
        for table in self.spec.keys():
            self.compiled[table] = self.compile_spec(
                table, self.spec[table], self.indices[table]
            )
            self.create_table(table)
        #
        # service tables
        #
        self.execute(
            '''
                     DROP TABLE IF EXISTS sources_options
                     '''
        )
        self.execute(
            '''
                     DROP TABLE IF EXISTS sources
                     '''
        )
        self.execute(
            '''
            DROP TABLE IF EXISTS config
        '''
        )
        self.execute(
            '''
            CREATE TABLE config
                (f_key TEXT PRIMARY KEY, f_value TEXT NOT NULL)
        '''
        )
        self.execute(
            '''
                     CREATE TABLE IF NOT EXISTS sources
                     (f_target TEXT PRIMARY KEY,
                      f_kind TEXT NOT NULL)
                     '''
        )
        self.execute(
            '''
                     CREATE TABLE IF NOT EXISTS sources_options
                     (f_target TEXT NOT NULL,
                      f_name TEXT NOT NULL,
                      f_type TEXT NOT NULL,
                      f_value TEXT NOT NULL,
                      FOREIGN KEY (f_target)
                          REFERENCES sources(f_target)
                          ON UPDATE CASCADE
                          ON DELETE CASCADE)
                     '''
        )
        for key, value in config.items():
            self.config[key] = value

    def merge_spec(self, table1, table2, table, schema_idx):
        spec1 = self.compiled[table1]
        spec2 = self.compiled[table2]
        names = spec1['names'] + spec2['names'][:-1]
        all_names = spec1['all_names'] + spec2['all_names'][2:-1]
        norm_names = spec1['norm_names'] + spec2['norm_names'][2:-1]
        idx = ('target', 'tflags') + schema_idx
        f_names = ['f_%s' % x for x in all_names]
        f_set = ['f_%s = %s' % (x, self.plch) for x in all_names]
        f_idx = ['f_%s' % x for x in idx]
        f_idx_match = ['%s.%s = %s' % (table2, x, self.plch) for x in f_idx]
        plchs = [self.plch] * len(f_names)
        return {
            'names': names,
            'all_names': all_names,
            'norm_names': norm_names,
            'idx': idx,
            'fnames': ','.join(f_names),
            'plchs': ','.join(plchs),
            'fset': ','.join(f_set),
            'knames': ','.join(f_idx),
            'fidx': ' AND '.join(f_idx_match),
        }

    def compile_spec(self, table, schema_names, schema_idx):
        # e.g.: index, flags, IFLA_IFNAME
        #
        names = []
        #
        # same + two internal fields
        #
        all_names = ['target', 'tflags']
        #
        #
        norm_names = ['target', 'tflags']

        bclass = self.classes.get(table)

        for name in schema_names:
            names.append(name[-1])
            all_names.append(name[-1])

            iclass = bclass
            if len(name) > 1:
                for step in name[:-1]:
                    imap = dict(iclass.nla_map)
                    iclass = getattr(iclass, imap[step])
            norm_names.append(iclass.nla2name(name[-1]))

        #
        # escaped names: f_index, f_flags, f_IFLA_IFNAME
        #
        # the reason: words like "index" are keywords in SQL
        # and we can not use them; neither can we change the
        # C structure
        #
        f_names = ['f_%s' % x for x in all_names]
        #
        # set the fields
        #
        # e.g.: f_flags = ?, f_IFLA_IFNAME = ?
        #
        # there are different placeholders:
        # ? -- SQLite3
        # %s -- PostgreSQL
        # so use self.plch here
        #
        f_set = ['f_%s = %s' % (x, self.plch) for x in all_names]
        #
        # the set of the placeholders to use in the INSERT statements
        #
        plchs = [self.plch] * len(f_names)
        #
        # the index schema; use target and tflags in every index
        #
        idx = ('target', 'tflags') + schema_idx
        #
        # the same, escaped: f_target, f_tflags etc.
        #
        f_idx = ['f_%s' % x for x in idx]
        #
        # normalized idx names
        #
        norm_idx = [iclass.nla2name(x) for x in idx]
        #
        # match the index fields, fully qualified
        #
        # interfaces.f_index = ?, interfaces.f_IFLA_IFNAME = ?
        #
        # the same issue with the placeholders
        #
        f_idx_match = ['%s.%s = %s' % (table, x, self.plch) for x in f_idx]

        return {
            'names': names,
            'all_names': all_names,
            'norm_names': norm_names,
            'idx': idx,
            'norm_idx': norm_idx,
            'fnames': ','.join(f_names),
            'plchs': ','.join(plchs),
            'fset': ','.join(f_set),
            'knames': ','.join(f_idx),
            'fidx': ' AND '.join(f_idx_match),
        }

    @publish
    def add_nl_source(self, target, kind, spec):
        '''
        A temprorary method, to be moved out
        '''
        # flush
        self.execute(
            '''
                DELETE FROM sources_options
                WHERE f_target = %s
            '''
            % self.plch,
            (target,),
        )
        self.execute(
            '''
                DELETE FROM sources
                WHERE f_target = %s
            '''
            % self.plch,
            (target,),
        )
        # add
        self.execute(
            '''
                INSERT INTO sources (f_target, f_kind)
                VALUES (%s, %s)
            '''
            % (self.plch, self.plch),
            (target, kind),
        )
        for key, value in spec.items():
            vtype = 'int' if isinstance(value, int) else 'str'
            self.execute(
                '''
                    INSERT INTO sources_options
                    (f_target, f_name, f_type, f_value)
                    VALUES (%s, %s, %s, %s)
                '''
                % (self.plch, self.plch, self.plch, self.plch),
                (target, key, vtype, value),
            )

    def execute(self, *argv, **kwarg):
        try:
            #
            # FIXME: add logging
            #
            for _ in range(MAX_ATTEMPTS):
                try:
                    self.cursor.execute(*argv, **kwarg)
                    break
                except (sqlite3.InterfaceError, sqlite3.OperationalError) as e:
                    self.log.debug('%s' % e)
                    #
                    # Retry on:
                    # -- InterfaceError: Error binding parameter ...
                    # -- OperationalError: SQL logic error
                    #
                    pass
            else:
                raise Exception('DB execute error: %s %s' % (argv, kwarg))
        except Exception:
            raise
        finally:
            self.connection.commit()  # no performance optimisation yet
        return self.cursor

    @publish
    def fetchone(self, *argv, **kwarg):
        for row in self.fetch(*argv, **kwarg):
            return row
        return None

    @publish
    def fetch(self, *argv, **kwarg):
        self.execute(*argv, **kwarg)
        while True:
            row_set = self.cursor.fetchmany()
            if not row_set:
                return
            for row in row_set:
                yield row

    @publish
    def backup(self, spec):
        if sys.version_info >= (3, 7) and self.provider == DBProvider.sqlite3:
            backup_connection = sqlite3.connect(spec)
            self.connection.backup(backup_connection)
            backup_connection.close()
        else:
            raise NotImplementedError()

    @publish
    def export(self, f='stdout'):
        close = False
        if f in ('stdout', 'stderr'):
            f = getattr(sys, f)
        elif isinstance(f, basestring):
            f = open(f, 'w')
            close = True
        try:
            for table in self.spec.keys():
                f.write('\ntable %s\n' % table)
                for record in self.execute('SELECT * FROM %s' % table):
                    f.write(' '.join([str(x) for x in record]))
                    f.write('\n')
                if self.rtnl_log:
                    f.write('\ntable %s_log\n' % table)
                    for record in self.execute('SELECT * FROM %s_log' % table):
                        f.write(' '.join([str(x) for x in record]))
                        f.write('\n')
        finally:
            if close:
                f.close()

    def close(self):
        if self.config['spec'] != ':memory:':
            # simply discard in-memory sqlite db on exit
            self.purge_snapshots()
            self.connection.commit()
        self.connection.close()

    @publish
    def commit(self):
        self.connection.commit()

    def create_table(self, table):
        req = ['f_target TEXT NOT NULL', 'f_tflags BIGINT NOT NULL DEFAULT 0']
        fields = []
        self.key_defaults[table] = {}
        for field in self.spec[table].items():
            #
            # Why f_?
            # 'Cause there are attributes like 'index' and such
            # names may not be used in SQL statements
            #
            field = (field[0][-1], field[1])
            fields.append('f_%s %s' % field)
            req.append('f_%s %s' % field)
            if field[1].strip().startswith('TEXT'):
                self.key_defaults[table][field[0]] = ''
            else:
                self.key_defaults[table][field[0]] = 0
        if table in self.foreign_keys:
            for key in self.foreign_keys[table]:
                spec = (
                    '(%s)' % ','.join(key['fields']),
                    '%s(%s)' % (key['parent'], ','.join(key['parent_fields'])),
                )
                req.append(
                    'FOREIGN KEY %s REFERENCES %s '
                    'ON UPDATE CASCADE '
                    'ON DELETE CASCADE ' % spec
                )
                #
                # make a unique index for compound keys on
                # the parent table
                #
                # https://sqlite.org/foreignkeys.html
                #
                if len(key['fields']) > 1:
                    idxname = 'uidx_%s_%s' % (
                        key['parent'],
                        '_'.join(key['parent_fields']),
                    )
                    self.execute(
                        'CREATE UNIQUE INDEX '
                        'IF NOT EXISTS %s ON %s' % (idxname, spec[1])
                    )

        req = ','.join(req)
        req = 'CREATE TABLE IF NOT EXISTS ' '%s (%s)' % (table, req)
        self.execute(req)

        index = ','.join(
            ['f_target', 'f_tflags']
            + ['f_%s' % x for x in self.indices[table]]
        )
        req = 'CREATE UNIQUE INDEX IF NOT EXISTS ' '%s_idx ON %s (%s)' % (
            table,
            table,
            index,
        )
        self.execute(req)

        #
        # create table for the transaction buffer: there go the system
        # updates while the transaction is not committed.
        #
        # w/o keys (yet)
        #
        # req = ['f_target TEXT NOT NULL',
        #        'f_tflags INTEGER NOT NULL DEFAULT 0']
        # req = ','.join(req)
        # self.execute('CREATE TABLE IF NOT EXISTS '
        #              '%s_buffer (%s)' % (table, req))
        #
        # create the log table, if required
        #
        if self.rtnl_log:
            req = [
                'f_tstamp BIGINT NOT NULL',
                'f_target TEXT NOT NULL',
                'f_event INTEGER NOT NULL',
            ] + fields
            req = ','.join(req)
            self.execute(
                'CREATE TABLE IF NOT EXISTS ' '%s_log (%s)' % (table, req)
            )

    def mark(self, target, mark):
        for table in self.spec:
            self.execute(
                '''
                         UPDATE %s SET f_tflags = %s
                         WHERE f_target = %s
                         '''
                % (table, self.plch, self.plch),
                (mark, target),
            )

    @publish
    def flush(self, target):
        for table in self.spec:
            self.execute(
                '''
                         DELETE FROM %s WHERE f_target = %s
                         '''
                % (table, self.plch),
                (target,),
            )

    @publish
    def save_deps(self, ctxid, weak_ref, iclass):
        uuid = uuid32()
        obj = weak_ref()
        obj_k = obj.key
        idx = self.indices[obj.table]
        conditions = []
        values = []
        for key in idx:
            conditions.append('f_%s = %s' % (key, self.plch))
            if key in obj_k:
                values.append(obj_k[key])
            else:
                values.append(obj.get(iclass.nla2name(key)))
        #
        # save the old f_tflags value
        #
        tflags = self.execute(
            '''
                           SELECT f_tflags FROM %s
                           WHERE %s
                           '''
            % (obj.table, ' AND '.join(conditions)),
            values,
        ).fetchone()[0]
        #
        # mark tflags for obj
        #
        obj.mark_tflags(uuid)

        #
        # f_tflags is used in foreign keys ON UPDATE CASCADE, so all
        # related records will be marked
        #
        for table in self.spec:
            self.log.debug('create snapshot %s_%s' % (table, ctxid))
            #
            # create the snapshot table
            #
            self.execute(
                '''
                         CREATE TABLE IF NOT EXISTS %s_%s
                         AS SELECT * FROM %s
                         WHERE
                             f_tflags IS NULL
                         '''
                % (table, ctxid, table)
            )
            #
            # copy the data -- is it possible to do it in one step?
            #
            self.execute(
                '''
                         INSERT INTO %s_%s
                         SELECT * FROM %s
                         WHERE
                             f_tflags = %s
                         '''
                % (table, ctxid, table, self.plch),
                [uuid],
            )
        #
        # unmark all the data
        #
        obj.mark_tflags(tflags)

        for table in self.spec:
            self.execute(
                '''
                         UPDATE %s_%s SET f_tflags = %s
                         '''
                % (table, ctxid, self.plch),
                [tflags],
            )
            self.snapshots['%s_%s' % (table, ctxid)] = weak_ref

    @publish
    def purge_snapshots(self):
        for table in tuple(self.snapshots):
            for _ in range(MAX_ATTEMPTS):
                try:
                    if self.provider == DBProvider.sqlite3:
                        self.execute('DROP TABLE %s' % table)
                    elif self.provider == DBProvider.psycopg2:
                        self.execute('DROP TABLE %s CASCADE' % table)
                    self.connection.commit()
                    del self.snapshots[table]
                    break
                except sqlite3.OperationalError:
                    #
                    # Retry on:
                    # -- OperationalError: database table is locked
                    #
                    time.sleep(random.random())
            else:
                raise Exception('DB snapshot error')

    @publish
    def get(self, table, spec):
        #
        # Retrieve info from the DB
        #
        # ndb.interfaces.get({'ifname': 'eth0'})
        #
        conditions = []
        values = []
        cls = self.classes[table]
        cspec = self.compiled[table]
        for key, value in spec.items():
            if key not in cspec['all_names']:
                key = cls.name2nla(key)
            if key not in cspec['all_names']:
                raise KeyError('field name not found')
            conditions.append('f_%s = %s' % (key, self.plch))
            values.append(value)
        req = 'SELECT * FROM %s WHERE %s' % (table, ' AND '.join(conditions))
        for record in self.fetch(req, values):
            yield dict(zip(self.compiled[table]['all_names'], record))

    def log_netlink(self, table, target, event, ctable=None):
        #
        # RTNL Logs
        #
        fkeys = self.compiled[table]['names']
        fields = ','.join(
            ['f_tstamp', 'f_target', 'f_event'] + ['f_%s' % x for x in fkeys]
        )
        pch = ','.join([self.plch] * (len(fkeys) + 3))
        values = [
            int(time.time() * 1000),
            target,
            event.get('header', {}).get('type', 0),
        ]
        for field in fkeys:
            value = event.get_attr(field) or event.get(field)
            if value is None and field in self.indices[ctable or table]:
                value = self.key_defaults[table][field]
            if isinstance(value, (dict, list, tuple, set)):
                value = json.dumps(value)
            values.append(value)
        self.execute(
            'INSERT INTO %s_log (%s) VALUES (%s)' % (table, fields, pch),
            values,
        )

    def load_netlink(self, table, target, event, ctable=None, propagate=False):
        #
        if self.rtnl_log:
            self.log_netlink(table, target, event, ctable)
        #
        # Update metrics
        #
        if 'stats' in event['header']:
            self.stats[target] = event['header']['stats']
        #
        # Periodic jobs
        #
        if time.time() - self.gctime > config.gc_timeout:
            self.gctime = time.time()

            # clean dead snapshots after GC timeout
            for name, wref in tuple(self.snapshots.items()):
                if wref() is None:
                    del self.snapshots[name]
                    try:
                        self.execute('DROP TABLE %s' % name)
                    except Exception as e:
                        self.log.debug(
                            'failed to remove table %s: %s' % (name, e)
                        )

            # clean marked routes
            self.execute(
                'DELETE FROM routes WHERE ' '(f_gc_mark + 5) < %s' % self.plch,
                (int(time.time()),),
            )
        #
        # The event type
        #
        if event['header'].get('type', 0) % 2:
            #
            # Delete an object
            #
            conditions = ['f_target = %s' % self.plch]
            values = [target]
            for key in self.indices[table]:
                conditions.append('f_%s = %s' % (key, self.plch))
                value = event.get(key) or event.get_attr(key)
                if value is None:
                    value = self.key_defaults[table][key]
                if isinstance(value, (dict, list, tuple, set)):
                    value = json.dumps(value)
                values.append(value)
            self.execute(
                'DELETE FROM %s WHERE'
                ' %s' % (table, ' AND '.join(conditions)),
                values,
            )
        else:
            #
            # Create or set an object
            #
            # field values
            values = [target, 0]
            # index values
            ivalues = [target, 0]
            compiled = self.compiled[table]
            # a map of sub-NLAs
            nodes = {}

            # fetch values (exc. the first two columns)
            for fname, ftype in self.spec[table].items():
                node = event

                # if the field is located in a sub-NLA
                if len(fname) > 1:
                    # see if we tried to get it already
                    if fname[:-1] not in nodes:
                        # descend
                        for steg in fname[:-1]:
                            node = node.get_attr(steg)
                            if node is None:
                                break
                        nodes[fname[:-1]] = node
                    # lookup the sub-NLA in the map
                    node = nodes[fname[:-1]]
                    # the event has no such sub-NLA
                    if node is None:
                        values.append(None)
                        continue

                # NLA have priority
                value = node.get_attr(fname[-1])
                if value is None:
                    value = node.get(fname[-1])
                if value is None and fname[-1] in self.compiled[table]['idx']:
                    value = self.key_defaults[table][fname[-1]]
                    node['attrs'].append((fname[-1], value))
                if isinstance(value, (dict, list, tuple, set)):
                    value = json.dumps(value)
                if fname[-1] in compiled['idx']:
                    ivalues.append(value)
                values.append(value)

            try:
                if self.provider == DBProvider.psycopg2:
                    #
                    # run UPSERT -- the DB provider must support it
                    #
                    (
                        self.execute(
                            'INSERT INTO %s (%s) VALUES (%s) '
                            'ON CONFLICT (%s) '
                            'DO UPDATE SET %s WHERE %s'
                            % (
                                table,
                                compiled['fnames'],
                                compiled['plchs'],
                                compiled['knames'],
                                compiled['fset'],
                                compiled['fidx'],
                            ),
                            (values + values + ivalues),
                        )
                    )
                    #
                elif self.provider == DBProvider.sqlite3:
                    #
                    # SQLite3 >= 3.24 actually has UPSERT, but ...
                    #
                    # We can not use here INSERT OR REPLACE as well, since
                    # it drops (almost always) records with foreign key
                    # dependencies. Maybe a bug in SQLite3, who knows.
                    #
                    count = (
                        self.execute(
                            '''
                                      SELECT count(*) FROM %s WHERE %s
                                      '''
                            % (table, compiled['fidx']),
                            ivalues,
                        ).fetchone()
                    )[0]
                    if count == 0:
                        self.execute(
                            '''
                                     INSERT INTO %s (%s) VALUES (%s)
                                     '''
                            % (table, compiled['fnames'], compiled['plchs']),
                            values,
                        )
                    else:
                        self.execute(
                            '''
                                     UPDATE %s SET %s WHERE %s
                                     '''
                            % (table, compiled['fset'], compiled['fidx']),
                            (values + ivalues),
                        )
                else:
                    raise NotImplementedError()
                #
            except Exception as e:
                #
                if propagate:
                    raise e
                #
                # A good question, what should we do here
                self.log.debug(
                    'load_netlink: %s %s %s' % (table, target, event)
                )
                self.log.error('load_netlink: %s' % traceback.format_exc())



import inspect
from unittest import mock

registry = []
use_mock = False


class LAB_API:
    def __init__(self, *argv, **kwarg):
        super().__init__(*argv, **kwarg)
        if use_mock:
            registry.append(self)
            for name, method in inspect.getmembers(
                self, predicate=inspect.ismethod
            ):
                setattr(self, name, mock.MagicMock(name=name, wraps=method))



import struct
from ctypes import (
    Structure,
    addressof,
    c_ubyte,
    c_uint,
    c_ushort,
    c_void_p,
    sizeof,
    string_at,
)
from socket import AF_PACKET, SOCK_RAW, SOL_SOCKET, errno, error, htons, socket

from pyroute2.iproute.linux import IPRoute

ETH_P_ALL = 3
SO_ATTACH_FILTER = 26
SO_DETACH_FILTER = 27


total_filter = [[0x06, 0, 0, 0]]


class sock_filter(Structure):
    _fields_ = [
        ('code', c_ushort),  # u16
        ('jt', c_ubyte),  # u8
        ('jf', c_ubyte),  # u8
        ('k', c_uint),
    ]  # u32


class sock_fprog(Structure):
    _fields_ = [('len', c_ushort), ('filter', c_void_p)]


def compile_bpf(code):
    ProgramType = sock_filter * len(code)
    program = ProgramType(*[sock_filter(*line) for line in code])
    sfp = sock_fprog(len(code), addressof(program[0]))
    return string_at(addressof(sfp), sizeof(sfp)), program


class RawSocket(socket):
    '''
    This raw socket binds to an interface and optionally installs a BPF
    filter.
    When created, the socket's buffer is cleared to remove packets that
    arrived before bind() or the BPF filter is installed.  Doing so
    requires calling recvfrom() which may raise an exception if the
    interface is down.
    In order to allow creating the socket when the interface is
    down, the ENETDOWN exception is caught and discarded.
    '''

    fprog = None

    def __init__(self, ifname, bpf=None):
        self.ifname = ifname
        # lookup the interface details
        with IPRoute() as ip:
            for link in ip.get_links():
                if link.get_attr('IFLA_IFNAME') == ifname:
                    break
            else:
                raise IOError(2, 'Link not found')
        self.l2addr = link.get_attr('IFLA_ADDRESS')
        self.ifindex = link['index']
        # bring up the socket
        socket.__init__(self, AF_PACKET, SOCK_RAW, htons(ETH_P_ALL))
        socket.bind(self, (self.ifname, ETH_P_ALL))
        if bpf:
            self.clear_buffer()
            fstring, self.fprog = compile_bpf(bpf)
            socket.setsockopt(self, SOL_SOCKET, SO_ATTACH_FILTER, fstring)
        else:
            self.clear_buffer(remove_total_filter=True)

    def clear_buffer(self, remove_total_filter=False):
        # there is a window of time after the socket has been created and
        # before bind/attaching a filter where packets can be queued onto the
        # socket buffer
        # see comments in function set_kernel_filter() in libpcap's
        # pcap-linux.c. libpcap sets a total filter which does not match any
        # packet.  It then clears what is already in the socket
        # before setting the desired filter
        total_fstring, prog = compile_bpf(total_filter)
        socket.setsockopt(self, SOL_SOCKET, SO_ATTACH_FILTER, total_fstring)
        self.setblocking(0)
        while True:
            try:
                self.recvfrom(0)
            except error as e:
                if e.args[0] == errno.ENETDOWN:
                    # we only get this exception once per down event
                    # there may be more packets left to clean
                    pass
                elif e.args[0] in [errno.EAGAIN, errno.EWOULDBLOCK]:
                    break
                else:
                    raise
        self.setblocking(1)
        if remove_total_filter:
            # total_fstring ignored
            socket.setsockopt(
                self, SOL_SOCKET, SO_DETACH_FILTER, total_fstring
            )

    def csum(self, data):
        if len(data) % 2:
            data += b'\x00'
        csum = sum(
            [
                struct.unpack('>H', data[x * 2 : x * 2 + 2])[0]
                for x in range(len(data) // 2)
            ]
        )
        csum = (csum >> 16) + (csum & 0xFFFF)
        csum += csum >> 16
        return ~csum & 0xFFFF



from pyroute2.netlink import NLA_F_NESTED, nla
from pyroute2.netlink.rtnl.tcmsg.common import tc_actions


class options(nla):
    nla_flags = NLA_F_NESTED
    nla_map = (
        ('TCA_GACT_UNSPEC', 'none'),
        ('TCA_GACT_TM', 'none'),
        ('TCA_GACT_PARMS', 'tca_gact_parms'),
        ('TCA_GACT_PROB', 'none'),
    )

    class tca_gact_parms(nla):
        fields = (
            ('index', 'I'),
            ('capab', 'I'),
            ('action', 'i'),
            ('refcnt', 'i'),
            ('bindcnt', 'i'),
        )


def get_parameters(kwarg):
    ret = {'attrs': []}
    a = tc_actions[kwarg.get('action', 'drop')]
    ret['attrs'].append(['TCA_GACT_PARMS', {'action': a}])
    return ret



import os
import time

from dateutil.parser import parse as isodate
from keystoneauth1 import session
from keystoneauth1.identity import v3
from keystoneclient.v3 import client as ksclient
from keystoneclient.v3.tokens import TokenManager


class OSAuthManager(object):
    def __init__(self, headers):
        # create a Keystone password object
        auth = v3.Password(
            auth_url=os.environ.get('OS_AUTH_URL'),
            username=os.environ.get('OS_USERNAME'),
            password=os.environ.get('OS_PASSWORD'),
            user_domain_name=(os.environ.get('OS_USER_DOMAIN_NAME')),
            project_id=os.environ.get('OS_PROJECT_ID'),
        )
        # create a session object
        sess = session.Session(auth=auth)
        # create a token manager
        tmanager = TokenManager(ksclient.Client(session=sess))
        # validate the token
        keystone_response = tmanager.validate(headers['X-Auth-Token'])
        # init attrs
        self.expire = isodate(keystone_response['expires_at']).timestamp()

    def check(self, obj, tag):
        if time.time() > self.expire:
            raise PermissionError('keystone token has been expired')
        return True



'''

.. testsetup::

    from pyroute2 import IPMock as IPRoute
    from pyroute2 import NDB
    from pyroute2 import config

    config.mock_iproute = True


.. testsetup:: preset_1

    from pyroute2 import NDB
    from pyroute2 import config

    config.mock_iproute = True
    ndb = NDB(
        sources=[
            {'target': 'localhost', 'kind': 'IPMock'},
            {'target': 'worker1.sample.com', 'kind': 'IPMock'},
            {'target': 'worker2.sample.com', 'kind': 'IPMock'},
        ]
    )

.. testsetup:: preset_br0_1

    from pyroute2 import NDB
    from pyroute2 import config
    config.mock_iproute = True
    ndb = NDB()
    ndb.interfaces.create(ifname='eth1', kind='dummy').commit()
    ndb.interfaces.create(ifname='br0', kind='bridge').commit()
    ndb.interfaces.create(ifname='bond0', kind='bond').commit()

.. testsetup:: preset_br0_2

    from pyroute2 import NDB
    from pyroute2 import config
    config.mock_iproute = True
    ndb = NDB()
    ndb.interfaces.create(ifname='br0', kind='bridge').commit()
    ndb.interfaces['br0'].add_port('eth0').commit()


List interfaces
===============

List interface keys:

.. testcode::

    with NDB(log='on') as ndb:
        for key in ndb.interfaces:
            print(key)


.. testoutput::
    :hide:

    ('localhost', 0, 0, 772, 1, 1, 0, '00:00:00:00:00:00', \
'00:00:00:00:00:00', 'lo', 65536, None, 'noqueue', None, 1000, 'UNKNOWN', 0, \
None, None, None, 0, None, 0, 1, 1, 1, 0, None, None, 0, 65535, 65536, None, \
None, None, 0, 0, None, None, None, None, None, None, 65536, None, None, \
'up', None, None, None, None, None, None, None, None, '[]')
    ('localhost', 0, 0, 772, 2, 1, 0, '52:54:00:72:58:b2', \
'ff:ff:ff:ff:ff:ff', 'eth0', 1500, None, 'fq_codel', None, 1000, 'UNKNOWN', \
0, None, None, None, 0, None, 0, 1, 1, 1, 0, None, None, 0, 65535, 65536, \
None, None, None, 0, 0, None, None, None, None, None, None, 65536, None, \
None, 'up', None, None, None, None, None, None, None, None, '[]')

NDB views support some dict methods: `items()`, `values()`, `keys()`:

.. testcode::

    with NDB(log='on') as ndb:
        for key, nic in ndb.interfaces.items():
            nic.set('state', 'up')
            nic.commit()

Get interface objects
=====================

The keys may be used as selectors to get interface objects:

.. testcode::

    with NDB() as ndb:
        for key in ndb.interfaces:
            print(ndb.interfaces[key])

.. testoutput::
    :hide:
    :options: +ELLIPSIS

    ...

Also possible selector formats are `dict()` and simple string. The latter
means the interface name:

.. testcode:: preset_1

    eth0 = ndb.interfaces['eth0']

Dict selectors are necessary to get interfaces by other properties:


.. testcode:: preset_1

    wrk1_eth0 = ndb.interfaces[{'target': 'worker1.sample.com',
                                'ifname': 'eth0'}]

    wrk2_eth0 = ndb.interfaces[{'target': 'worker2.sample.com',
                                'address': '52:54:00:72:58:b2'}]

Change nic properties
=====================

Changing MTU and MAC address:

.. testcode:: preset_1

    with ndb.interfaces['eth0'] as eth0:
        eth0['mtu'] = 1248
        eth0['address'] = '00:11:22:33:44:55'
    # --> <-- eth0.commit() is called by the context manager

One can change a property either using the assignment statement, or
using the `.set()` routine:

.. testcode:: preset_1

    # same code
    with ndb.interfaces['eth0'] as eth0:
        eth0.set('mtu', 1248)
        eth0.set('address', '00:11:22:33:44:55')


Create virtual interfaces
=========================

Create a bridge and add a port, `eth0`:

.. testcode:: preset_1

    with ndb.interfaces.create(ifname='br0', kind='bridge') as br0:
        br0.add_port('eth0')

Bridge and bond ports
=====================

Add bridge and bond ports one can use specific API:

.. testcode:: preset_br0_1

    with ndb.interfaces['br0'] as br0:
        br0.add_port('eth0')
        br0.add_port('eth1')
        br0.set('br_max_age', 1024)
        br0.set('br_forward_delay', 1500)

    with ndb.interfaces['bond0'] as bond0:
        bond0.add_port('eth0')
        bond0.add_port('eth1')

To remove a port:

.. testcode:: preset_br0_2

    with ndb.interfaces['br0'] as br0:
        br0.del_port('eth0')

Or by setting the master property on a port, in the same
way as with `IPRoute`:

.. testcode:: preset_br0_1

    index = ndb.interfaces['br0']['index']

    # add a port to a bridge
    with ndb.interfaces['eth0'] as eth0:
        eth0.set('master', index)

    # remove a port from a bridge
    with ndb.interfaces['eth0'] as eth0:
        eth0.set('master', 0)
'''

import errno
import json
import traceback

from pyroute2.common import basestring
from pyroute2.config import AF_BRIDGE
from pyroute2.netlink.exceptions import NetlinkError
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
from pyroute2.netlink.rtnl.p2pmsg import p2pmsg
from pyroute2.requests.link import LinkFieldFilter

from ..auth_manager import AuthManager, check_auth
from ..objects import RTNL_Object


def load_ifinfmsg(schema, target, event):
    #
    # link goes down: flush all related routes
    #
    if not event['flags'] & 1:
        schema.execute(
            'DELETE FROM routes WHERE '
            'f_target = %s AND '
            'f_RTA_OIF = %s OR f_RTA_IIF = %s'
            % (schema.plch, schema.plch, schema.plch),
            (target, event['index'], event['index']),
        )
    #
    # ignore wireless updates
    #
    if event.get_attr('IFLA_WIRELESS'):
        return
    #
    # IFLA_PROP_LIST, IFLA_ALT_IFNAME
    #
    prop_list = event.get('IFLA_PROP_LIST')
    event['alt_ifname_list'] = []
    if prop_list is not None:
        for ifname in prop_list.altnames():
            event['alt_ifname_list'].append(ifname)

    #
    # AF_BRIDGE events
    #
    if event['family'] == AF_BRIDGE:
        #
        schema.load_netlink('af_bridge_ifs', target, event)
        try:
            vlans = event.get_attr('IFLA_AF_SPEC').get_attrs(
                'IFLA_BRIDGE_VLAN_INFO'
            )
        except AttributeError:
            # AttributeError: 'NoneType' object has no attribute 'get_attrs'
            # -- vlan filters not supported
            return

        # flush the old vlans info
        schema.execute(
            '''
                       DELETE FROM af_bridge_vlans
                       WHERE
                           f_target = %s
                           AND f_index = %s
                       '''
            % (schema.plch, schema.plch),
            (target, event['index']),
        )
        for v in vlans:
            v['index'] = event['index']
            v['header'] = {'type': event['header']['type']}
            schema.load_netlink('af_bridge_vlans', target, v)

        return

    schema.load_netlink('interfaces', target, event)
    #
    # load ifinfo, if exists
    #
    if not event['header'].get('type', 0) % 2:
        linkinfo = event.get_attr('IFLA_LINKINFO')
        if linkinfo is not None:
            iftype = linkinfo.get_attr('IFLA_INFO_KIND')
            table = 'ifinfo_%s' % iftype
            if iftype == 'gre':
                ifdata = linkinfo.get_attr('IFLA_INFO_DATA')
                local = ifdata.get_attr('IFLA_GRE_LOCAL')
                remote = ifdata.get_attr('IFLA_GRE_REMOTE')
                p2p = p2pmsg()
                p2p['index'] = event['index']
                p2p['family'] = 2
                p2p['attrs'] = [('P2P_LOCAL', local), ('P2P_REMOTE', remote)]
                schema.load_netlink('p2p', target, p2p)
            elif iftype == 'veth':
                link = event.get_attr('IFLA_LINK')
                ifname = event.get_attr('IFLA_IFNAME')
                # for veth interfaces, IFLA_LINK points to
                # the peer -- but NOT in automatic updates
                if (not link) and (
                    (target,) in schema.fetch('SELECT f_target FROM SOURCES')
                ):
                    schema.log.debug('reload veth %s' % event['index'])
                    try:
                        update = schema.sources[target].api(
                            'link', 'get', index=event['index']
                        )
                        update = tuple(update)[0]
                        return schema.load_netlink(
                            'interfaces', target, update
                        )
                    except NetlinkError as e:
                        if e.code == errno.ENODEV:
                            schema.log.debug(f"interface has gone: {ifname}")

            if table in schema.spec:
                ifdata = linkinfo.get_attr('IFLA_INFO_DATA')
                if ifdata is not None:
                    ifdata['header'] = {}
                    ifdata['index'] = event['index']
                    schema.load_netlink(table, target, ifdata)


ip_tunnels = ('gre', 'gretap', 'ip6gre', 'ip6gretap', 'ip6tnl', 'sit', 'ipip')

schema_ifinfmsg = (
    ifinfmsg.sql_schema().push('alt_ifname_list', 'TEXT').unique_index('index')
)

schema_brinfmsg = (
    ifinfmsg.sql_schema()
    .unique_index('index')
    .foreign_key(
        'interface',
        ('f_target', 'f_tflags', 'f_index'),
        ('f_target', 'f_tflags', 'f_index'),
    )
)

schema_p2pmsg = (
    p2pmsg.sql_schema()
    .unique_index('index')
    .foreign_key(
        'interfaces',
        ('f_target', 'f_tflags', 'f_index'),
        ('f_target', 'f_tflags', 'f_index'),
    )
)

schema_af_bridge_vlans = (
    ifinfmsg.af_spec_bridge.vlan_info.sql_schema()
    .push('index', 'INTEGER')
    .unique_index('vid', 'index')
    .foreign_key(
        'af_bridge_ifs',
        ('f_target', 'f_tflags', 'f_index'),
        ('f_target', 'f_tflags', 'f_index'),
    )
)

init = {
    'specs': [
        ['interfaces', schema_ifinfmsg],
        ['af_bridge_ifs', schema_ifinfmsg],
        ['af_bridge_vlans', schema_af_bridge_vlans],
        ['p2p', schema_p2pmsg],
    ],
    'classes': [
        ['interfaces', ifinfmsg],
        ['af_bridge_ifs', ifinfmsg],
        ['vlans', ifinfmsg],
        ['af_bridge_vlans', ifinfmsg.af_spec_bridge.vlan_info],
        ['p2p', p2pmsg],
    ],
    'event_map': {ifinfmsg: [load_ifinfmsg]},
}

ifinfo_names = (
    'bridge',
    'bond',
    'vlan',
    'vxlan',
    'gre',
    'gretap',
    'ip6gre',
    'ip6gretap',
    'ip6tnl',
    'ipip',
    'ipvlan',
    'sit',
    'macvlan',
    'macvtap',
    'tun',
    'vrf',
    'vti',
    'vti6',
)
supported_ifinfo = {x: ifinfmsg.ifinfo.data_map[x] for x in ifinfo_names}
#
# load supported ifinfo
#
for name, data in supported_ifinfo.items():
    name = 'ifinfo_%s' % name
    init['classes'].append([name, data])
    schema = (
        data.sql_schema()
        .push('index', 'BIGINT')
        .unique_index('index')
        .foreign_key(
            'interfaces',
            ('f_target', 'f_tflags', 'f_index'),
            ('f_target', 'f_tflags', 'f_index'),
        )
    )
    init['specs'].append([name, schema])


def _cmp_master(self, value):
    if self['master'] == value:
        return True
    elif self['master'] == 0 and value is None:
        dict.__setitem__(self, 'master', None)
        return True
    return False


class Vlan(RTNL_Object):
    table = 'af_bridge_vlans'
    msg_class = ifinfmsg.af_spec_bridge.vlan_info
    api = 'vlan_filter'

    @classmethod
    def _count(cls, view):
        if view.chain:
            return view.ndb.task_manager.db_fetchone(
                'SELECT count(*) FROM %s WHERE f_index = %s'
                % (view.table, view.ndb.schema.plch),
                [view.chain['index']],
            )
        else:
            return view.ndb.task_manager.db_fetchone(
                'SELECT count(*) FROM %s' % view.table
            )

    @classmethod
    def _dump_where(cls, view):
        if view.chain:
            plch = view.ndb.schema.plch
            where = '''
                    WHERE
                        main.f_target = %s AND
                        main.f_index = %s
                    ''' % (
                plch,
                plch,
            )
            values = [view.chain['target'], view.chain['index']]
        else:
            where = ''
            values = []
        return (where, values)

    @classmethod
    def summary(cls, view):
        req = '''
              SELECT
                  main.f_target, main.f_tflags, main.f_vid,
                  intf.f_IFLA_IFNAME
              FROM
                  af_bridge_vlans AS main
              INNER JOIN
                  interfaces AS intf
              ON
                  main.f_index = intf.f_index
                  AND main.f_target = intf.f_target
              '''
        yield ('target', 'tflags', 'vid', 'ifname')
        where, values = cls._dump_where(view)
        for record in view.ndb.task_manager.db_fetch(req + where, values):
            yield record

    @staticmethod
    def compare_record(left, right):
        if isinstance(right, int):
            return right == left['vid']

    def __init__(self, *argv, **kwarg):
        kwarg['iclass'] = ifinfmsg.af_spec_bridge.vlan_info
        if 'auth_managers' not in kwarg or kwarg['auth_managers'] is None:
            kwarg['auth_managers'] = []
        log = argv[0].ndb.log.channel('vlan auth')
        kwarg['auth_managers'].append(
            AuthManager(
                {'obj:read': True, 'obj:list': True, 'obj:modify': False}, log
            )
        )
        super(Vlan, self).__init__(*argv, **kwarg)

    def make_req(self, prime):
        ret = {}
        if 'index' in self:
            ret['index'] = self['index']
        ret['vlan_info'] = {'vid': self['vid']}
        if 'flags' in self:
            ret['vlan_info']['flags'] = self['flags']
        return ret

    def make_idx_req(self, prime):
        return self.make_req(prime)


class Interface(RTNL_Object):
    table = 'interfaces'
    msg_class = ifinfmsg
    api = 'link'
    key_extra_fields = ['IFLA_IFNAME']
    resolve_fields = ['vxlan_link', 'link', 'master']
    fields_cmp = {'master': _cmp_master}
    fields_load_transform = {
        'alt_ifname_list': lambda x: list(json.loads(x or '[]'))
    }
    field_filter = LinkFieldFilter

    @classmethod
    def _count(cls, view):
        if view.chain:
            return view.ndb.task_manager.db_fetchone(
                'SELECT count(*) FROM %s WHERE f_IFLA_MASTER = %s'
                % (view.table, view.ndb.schema.plch),
                [view.chain['index']],
            )
        else:
            return view.ndb.task_manager.db_fetchone(
                'SELECT count(*) FROM %s' % view.table
            )

    @classmethod
    def _dump_where(cls, view):
        if view.chain:
            plch = view.ndb.schema.plch
            where = '''
                    WHERE
                        f_target = %s AND
                        f_IFLA_MASTER = %s
                    ''' % (
                plch,
                plch,
            )
            values = [view.chain['target'], view.chain['index']]
        else:
            where = 'WHERE f_index != 0'
            values = []
        return (where, values)

    @classmethod
    def summary(cls, view):
        req = '''
              SELECT
                  f_target, f_tflags, f_index,
                  f_IFLA_IFNAME, f_IFLA_ADDRESS,
                  f_flags, f_IFLA_INFO_KIND
              FROM
                  interfaces
              '''
        yield (
            'target',
            'tflags',
            'index',
            'ifname',
            'address',
            'flags',
            'kind',
        )
        where, values = cls._dump_where(view)
        for record in view.ndb.task_manager.db_fetch(req + where, values):
            yield record

    def mark_tflags(self, mark):
        plch = (self.schema.plch,) * 3
        self.schema.execute(
            '''
                            UPDATE interfaces SET
                                f_tflags = %s
                            WHERE f_index = %s AND f_target = %s
                            '''
            % plch,
            (mark, self['index'], self['target']),
        )

    def __init__(self, *argv, **kwarg):
        kwarg['iclass'] = ifinfmsg
        self.event_map = {ifinfmsg: "load_rtnlmsg"}
        self._alt_ifname_orig = set()
        dict.__setitem__(self, 'alt_ifname_list', list())
        dict.__setitem__(self, 'state', 'unknown')
        warnings = []
        if isinstance(argv[1], dict):
            if 'reuse' in argv[1]:
                warnings.append('ignore IPDB-specific `reuse` keyword')
                del argv[1]['reuse']
            if argv[1].get('create') and 'ifname' not in argv[1]:
                raise Exception('specify at least ifname')
            # type specific cases
            if argv[1].get('kind') == 'tuntap':
                # translate custom tuntap format into the native tun
                warnings.append('translated tuntap ifinfo into tun, no flags')
                argv[1]['kind'] = 'tun'
                if argv[1].get('mode') == 'tun':
                    argv[1]['tun_type'] = 1
                elif argv[1].get('mode') == 'tap':
                    argv[1]['tun_type'] = 2
                else:
                    raise TypeError('tun type error')
                del argv[1]['mode']
                if 'uid' in argv[1]:
                    argv[1]['tun_owner'] = argv[1].pop('uid')
                if 'gid' in argv[1]:
                    argv[1]['tun_owner'] = argv[1].pop('gid')
        super(Interface, self).__init__(*argv, **kwarg)
        for line in warnings:
            self.log.warning(line)

    @property
    def ipaddr(self):
        return self.view.ndb._get_view('addresses', chain=self)

    @property
    def ports(self):
        return self.view.ndb._get_view('interfaces', chain=self)

    @property
    def routes(self):
        return self.view.ndb._get_view('routes', chain=self)

    @property
    def neighbours(self):
        return self.view.ndb._get_view('neighbours', chain=self)

    @property
    def vlans(self):
        return self.view.ndb._get_view('af_bridge_vlans', chain=self)

    @property
    def context(self):
        ctx = {}
        if self.get('target'):
            ctx['target'] = self['target']
        if self.get('index'):
            ctx['index'] = self['index']
        return ctx

    @classmethod
    def compare_record(self, left, right):
        # specific compare
        if isinstance(right, basestring):
            return right == left['ifname'] or right == left['address']

    @check_auth('obj:modify')
    def add_vlan(self, spec):
        def do_add_vlan(self, mode, spec):
            try:
                method = getattr(self.vlan.create(spec), mode)
                return [method()]
            except Exception as e_s:
                e_s.trace = traceback.format_stack()
                return [e_s]

        self._apply_script.append((do_add_vlan, {'spec': spec}))
        return self

    @check_auth('obj:modify')
    def del_vlan(self, spec):
        def do_del_vlan(self, mode, spec):
            try:
                method = getattr(self.vlan[spec].remove(), mode)
                return [method()]
            except Exception as e_s:
                e_s.trace = traceback.format_stack()
                return [e_s]

        self._apply_script.append((do_del_vlan, {'spec': spec}))
        return self

    @check_auth('obj:modify')
    def add_neighbour(self, spec=None, **kwarg):
        spec = spec or kwarg

        def do_add_neighbour(self, mode, spec):
            try:
                method = getattr(self.neighbours.create(spec), mode)
                return [method()]
            except Exception as e_s:
                e_s.trace = traceback.format_stack()
                return [e_s]

        self._apply_script.append((do_add_neighbour, {'spec': spec}))
        return self

    @check_auth('obj:modify')
    def del_neighbour(self, spec=None, **kwarg):
        spec = spec or dict(kwarg)

        def do_del_neighbour(self, mode, spec):
            ret = []
            if isinstance(spec, basestring):
                specs = [spec]
            elif callable(spec):
                specs = self.ipaddr.dump()
                specs.select_records(spec)
            else:
                specs = self.ipaddr.dump()
                specs.select_records(**spec)
            for sp in specs:
                try:
                    method = getattr(self.neighbours.locate(sp).remove(), mode)
                    ret.append(method())
                except KeyError:
                    pass
                except Exception as e_s:
                    e_s.trace = traceback.format_stack()
                    ret.append(e_s)
            if not ret:
                ret = KeyError('no neighbour records matched')
            return ret

        self._apply_script.append((do_del_neighbour, {'spec': spec}))
        return self

    @check_auth('obj:modify')
    def add_ip(self, spec=None, **kwarg):
        spec = spec or kwarg

        def do_add_ip(self, mode, spec):
            try:
                method = getattr(self.ipaddr.create(spec), mode)
                return [method()]
            except Exception as e_s:
                e_s.trace = traceback.format_stack()
                return [e_s]

        self._apply_script.append((do_add_ip, {'spec': spec}))
        return self

    @check_auth('obj:modify')
    def del_ip(self, spec=None, **kwarg):
        spec = spec or kwarg

        def do_del_ip(self, mode, spec):
            ret = []
            if isinstance(spec, basestring):
                specs = [spec]
            elif callable(spec):
                specs = self.ipaddr.dump()
                specs.select_records(spec)
            else:
                specs = self.ipaddr.dump()
                specs.select_records(**spec)
            for sp in specs:
                try:
                    method = getattr(self.ipaddr.locate(sp).remove(), mode)
                    ret.append(method())
                except KeyError:
                    pass
                except Exception as e_s:
                    e_s.trace = traceback.format_stack()
                    ret.append(e_s)
            if not ret:
                ret = KeyError('no address records matched')
            return ret

        self._apply_script.append((do_del_ip, {'spec': spec}))
        return self

    @check_auth('obj:modify')
    def add_port(self, spec):
        def do_add_port(self, mode, spec):
            try:
                port = self.view[spec]
                if port['target'] != self['target']:
                    raise ValueError('target must be the same')
                port['master'] = self['index']
                getattr(port, mode)()
                return [port]
            except Exception as e_s:
                e_s.trace = traceback.format_stack()
                return [e_s]

        self._apply_script.append((do_add_port, {'spec': spec}))
        return self

    @check_auth('obj:modify')
    def del_port(self, spec):
        def do_del_port(self, mode, spec):
            try:
                port = self.view[spec]
                if port['master'] != self['index']:
                    raise ValueError('wrong port master index')
                if port['target'] != self['target']:
                    raise ValueError('target must be the same')
                port['master'] = 0
                getattr(port, mode)()
                return [port]
            except Exception as e_s:
                e_s.trace = traceback.format_stack()
                return [e_s]

        self._apply_script.append((do_del_port, {'spec': spec}))
        return self

    @check_auth('obj:modify')
    def add_altname(self, ifname):
        new_list = set(self['alt_ifname_list'])
        new_list.add(ifname)
        self['alt_ifname_list'] = list(new_list)

    @check_auth('obj:modify')
    def del_altname(self, ifname):
        new_list = set(self['alt_ifname_list'])
        new_list.remove(ifname)
        self['alt_ifname_list'] = list(new_list)

    @check_auth('obj:modify')
    def __setitem__(self, key, value):
        if key == 'peer':
            dict.__setitem__(self, key, value)
        elif key == 'target' and self.state == 'invalid':
            dict.__setitem__(self, key, value)
        elif key == 'net_ns_fd' and self.state == 'invalid':
            dict.__setitem__(self, 'target', value)
        elif (
            key == 'target' and self.get('target') and self['target'] != value
        ):
            super(Interface, self).__setitem__('net_ns_fd', value)
        else:
            super(Interface, self).__setitem__(key, value)

    @classmethod
    def spec_normalize(cls, processed, spec):
        '''
        Interface key normalization::

            { ... }  ->  { ... }
            "eth0"   ->  {"ifname": "eth0", ...}
            1        ->  {"index": 1, ...}

        '''
        if isinstance(spec, basestring):
            processed['ifname'] = spec
        elif isinstance(spec, int):
            processed['index'] = spec
        return processed

    def complete_key(self, key):
        if isinstance(key, dict):
            ret_key = key
        else:
            ret_key = {'target': self.ndb.localhost}
        if isinstance(key, basestring):
            ret_key['ifname'] = key
        elif isinstance(key, int):
            ret_key['index'] = key
        return super(Interface, self).complete_key(ret_key)

    def is_peer(self, other):
        '''Evaluate whether the given interface "points at" this one.'''
        if other['kind'] == 'vlan':
            return (
                other['target'] == self['target']
                and other['link'] == self['index']
            )

        elif other['kind'] == 'vxlan':
            return (
                other['target'] == self['target']
                and other['vxlan_link'] == self['index']
            )

        elif other['kind'] == self['kind'] == 'veth':
            other_link = other.get('link')

            if other_link != self['index']:
                return False

            other_link_netnsid = other.get('link_netnsid')
            if other_link_netnsid is not None:
                self_source = self.sources[self['target']]
                other_source = other.sources[other['target']]
                info = other_source.api(
                    'get_netnsid',
                    pid=self_source.api('get_pid'),
                    target_nsid=other_link_netnsid,
                )
                return info['current_nsid'] == other_link_netnsid

            return self['target'] == other['target']

    def set_xdp_fd(self, fd):
        self.sources[self['target']].api(
            'link', 'set', index=self['index'], xdp_fd=fd
        )

    def snapshot(self, ctxid=None):
        # 1. make own snapshot
        snp = super(Interface, self).snapshot(ctxid=ctxid)
        # 2. collect dependencies and store in self.snapshot_deps
        for spec in self.ndb.interfaces.getmany(
            {'IFLA_MASTER': self['index']}
        ):
            # bridge ports
            link = type(self)(
                self.view, spec, auth_managers=self.auth_managers
            )
            snp.snapshot_deps.append((link, link.snapshot()))
        for spec in self.ndb.interfaces.getmany({'IFLA_LINK': self['index']}):
            link = type(self)(
                self.view, spec, auth_managers=self.auth_managers
            )
            # vlans & veth
            if self.is_peer(link) and not link.is_peer(self):
                snp.snapshot_deps.append((link, link.snapshot()))
        # return the root node
        return snp

    def make_req(self, prime):
        req = super(Interface, self).make_req(prime)
        #
        # --> link('set', ...)
        if self.state == 'system':
            req['master'] = self['master']
            #
            # FIXME: make type plugins?
            kind = self['kind']
            if kind in ip_tunnels:
                req['kind'] = kind
                for key in self:
                    if (
                        key.startswith(f'{kind}_')
                        and key not in req
                        and self[key]
                    ):
                        req[key] = self[key]
        return req

    @check_auth('obj:modify')
    def apply_altnames(self, alt_ifname_setup):
        alt_ifname_remove = set(self['alt_ifname_list']) - alt_ifname_setup
        alt_ifname_add = alt_ifname_setup - set(self['alt_ifname_list'])
        for ifname in alt_ifname_remove:
            self.sources[self['target']].api(
                'link', 'property_del', index=self['index'], altname=ifname
            )
        for ifname in alt_ifname_add:
            self.sources[self['target']].api(
                'link', 'property_add', index=self['index'], altname=ifname
            )
        self.load_from_system()
        self.load_sql(set_state=False)
        if set(self['alt_ifname_list']) != alt_ifname_setup:
            raise Exception('could not setup alt ifnames')

    @check_auth('obj:modify')
    def apply(self, rollback=False, req_filter=None, mode='apply'):
        # translate string link references into numbers
        for key in ('link', 'master'):
            if key in self and isinstance(self[key], basestring):
                self[key] = self.ndb.interfaces[self[key]]['index']
        setns = self.state.get() == 'setns'
        remove = self.state.get() == 'remove'
        alt_ifname_setup = set(self['alt_ifname_list'])
        if 'alt_ifname_list' in self.changed:
            self.changed.remove('alt_ifname_list')
        try:
            super(Interface, self).apply(rollback, req_filter, mode)
            if setns:
                self.load_value('target', self['net_ns_fd'])
                dict.__setitem__(self, 'net_ns_fd', None)
                spec = self.load_sql()
                if spec:
                    self.state.set('system')
            if not remove:
                self.apply_altnames(alt_ifname_setup)

        except NetlinkError as e:
            if (
                e.code == 95
                and self.get('master') is not None
                and self.get('master') > 0
                and self.state == 'invalid'
            ):
                #
                # on some old kernels it is impossible to create
                # interfaces with master set; attempt to do it in
                # two steps
                def req_filter(req):
                    return dict(
                        [
                            x
                            for x in req.items()
                            if not x[0].startswith('master')
                        ]
                    )

                self.apply(rollback, req_filter, mode)
                self.apply(rollback, None, mode)

            elif (
                e.code == 95
                and self.get('br_vlan_filtering') is not None
                and self.get('br_vlan_filtering') == 0
            ):
                #
                # if vlan filtering is not enabled, then the parameter
                # is reported by netlink, but not accepted upon bridge
                # creation, so simply strip it
                def req_filter(req):
                    return dict(
                        [
                            x
                            for x in req.items()
                            if not x[0].startswith('br_vlan_')
                        ]
                    )

                self.apply(rollback, req_filter, mode)
            else:
                raise
        if ('net_ns_fd' in self.get('peer', {})) and (
            self['peer']['net_ns_fd'] in self.view.ndb.sources
        ):
            # wait for the peer in net_ns_fd, only if the netns
            # is connected to the NDB instance
            self.view.wait(
                target=self['peer']['net_ns_fd'],
                ifname=self['peer']['ifname'],
                timeout=5,
            )
        return self

    def hook_apply(self, method, **spec):
        if method == 'set':
            if self['kind'] == 'bridge':
                keys = filter(lambda x: x.startswith('br_'), self.changed)
                if keys:
                    req = {
                        'index': self['index'],
                        'kind': 'bridge',
                        'family': AF_BRIDGE,
                    }
                    for key in keys:
                        req[key] = self[key]
                    self.sources[self['target']].api(self.api, method, **req)
                    # FIXME: make a reasonable shortcut for this
                    self.load_from_system()
            elif self['kind'] in ip_tunnels and self['state'] == 'down':
                # force reading attributes for tunnels in the down state
                self.load_from_system()
        elif method == 'add':
            if self['kind'] == 'tun':
                self.load_sql()
                self.load_event.wait(0.1)
                if 'index' not in self:
                    raise NetlinkError(errno.EAGAIN)
                update = self.sources[self['target']].api(
                    self.api, 'get', index=self['index']
                )
                self.ndb._event_queue.put(update)

    def load_from_system(self):
        (
            self.ndb._event_queue.put(
                self.sources[self['target']].api(
                    self.api, 'get', index=self['index']
                )
            )
        )

    def load_sql(self, *argv, **kwarg):
        spec = super(Interface, self).load_sql(*argv, **kwarg)
        if spec:
            tname = 'ifinfo_%s' % self['kind']
            if tname in self.schema.compiled:
                names = self.schema.compiled[tname]['norm_names']
                spec = self.ndb.task_manager.db_fetchone(
                    'SELECT * from %s WHERE f_index = %s'
                    % (tname, self.schema.plch),
                    (self['index'],),
                )
                if spec:
                    self.update(dict(zip(names, spec)))
        return spec

    def load_rtnlmsg(self, *argv, **kwarg):
        super(Interface, self).load_rtnlmsg(*argv, **kwarg)

    def key_repr(self):
        return '%s/%s' % (
            self.get('target', ''),
            self.get('ifname', self.get('index', '')),
        )
